# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests the ability to connect to the SAPI server with the `dwave.cloud.qpu.Client`.

test_mock_solver_loading.py duplicates some of these tests against a mock server.
"""

import json
import time
import warnings
import unittest
from unittest import mock
from contextlib import contextmanager

import requests
from plucky import merge

from dwave.cloud.api import constants
from dwave.cloud.client import Client
from dwave.cloud.solver import StructuredSolver, UnstructuredSolver
from dwave.cloud.exceptions import (
    SolverAuthenticationError, SolverError, SolverNotFoundError)
from dwave.cloud.testing import iterable_mock_open, isolated_environ
import dwave.cloud

from tests import config


DEFAULT_REGIONS = {
    constants.DEFAULT_REGION: dict(
        endpoint=constants.DEFAULT_SOLVER_API_ENDPOINT)}

def get_default_regions(*args, **kwargs):
    return DEFAULT_REGIONS


@unittest.skipUnless(config, "No live server configuration available.")
class ConnectivityTests(unittest.TestCase):
    """Test connecting and related failure modes."""

    def test_bad_url(self):
        """Connect with a bad URL."""
        with self.assertRaises(IOError):
            invalid_config = config.copy()
            invalid_config.update(endpoint='invalid-endpoint')
            with Client(**invalid_config) as client:
                client.get_solvers()

    def test_bad_token(self):
        """Connect with a bad token."""
        with self.assertRaises(SolverAuthenticationError):
            invalid_config = config.copy()
            invalid_config.update(token='invalid-token')
            with Client(**invalid_config) as client:
                client.get_solvers()

    def test_good_connection(self):
        """Connect with valid connection settings (url/token/proxy/etc)."""
        with Client(**config) as client:
            self.assertTrue(len(client.get_solvers()) > 0)


@unittest.skipUnless(config, "No live server configuration available.")
class TestTimeouts(unittest.TestCase):
    """Test timeout works for all Client connections."""

    def test_request_timeout(self):
        with self.assertRaises(dwave.cloud.exceptions.RequestTimeout):
            with Client(request_timeout=0.00001, **config) as client:
                client.get_solvers()

    def test_polling_timeout(self):
        with self.assertRaises(dwave.cloud.exceptions.PollingTimeout):
            with Client(polling_timeout=0.00001, **config) as client:
                solver = client.get_solver()
                solver.sample_qubo({next(iter(solver.edges)): 0}).result()


@unittest.skipUnless(config, "No live server configuration available.")
class TestRetrieveAnswer(unittest.TestCase):
    """Test loading a problem from its id"""

    def test_retrieve_answer(self):
        """Answer retrieved based on problem_id in a new client."""

        with Client(**config) as client:
            solver = client.get_solver()

            h = {v: -1 for v in solver.nodes}

            f = solver.sample_ising(h, {})

            # the id is not set right away
            while f.id is None:
                time.sleep(.01)

            id_ = f.id
        
        with Client(**config) as client:
            # get a "new" client
            f2 = client.retrieve_answer(id_)

            self.assertIn('solutions', f2.result())

    def test_retrieve_invalid_answer_id(self):
        """Loading result fails when problem_id invalid."""

        with Client(**config) as client:
            f = client.retrieve_answer('invalid_id')
            with self.assertRaises(SolverError):
                f.result()

@unittest.skipUnless(config, "No live server configuration available.")
class SolverLoading(unittest.TestCase):
    """Test loading solvers in a few different configurations."""

    def test_list_all_solvers(self):
        """List all the solvers."""
        with Client(**config) as client:
            self.assertTrue(len(client.get_solvers()) > 0)

    def test_load_all_solvers(self):
        """List and retrieve all the solvers."""
        with Client(**config) as client:
            for solver in client.get_solvers():
                self.assertEqual(client.get_solver(name=solver.id).id, solver.id)

    def test_load_bad_solvers(self):
        """Try to load a nonexistent solver."""
        with Client(**config) as client:
            with self.assertRaises(SolverNotFoundError):
                client.get_solver("not-a-solver")

    def test_load_any_solver(self):
        """Load a single solver without calling get_solvers."""
        with Client(**config) as client:
            self.assertTrue(client.get_solver(software=True).software)

    def test_get_solver_no_defaults(self):
        """Specialized client returns the correct solver by default."""
        from dwave.cloud import qpu, sw, hybrid

        conf = config.copy()
        del conf['solver']

        with Client(**conf) as base_client:

            with qpu.Client(**conf) as client:
                solvers = {s.id for s in base_client.get_solvers(qpu=True)}
                if solvers:
                    self.assertIn(client.get_solver().id, solvers)
                else:
                    self.assertRaises(SolverError, client.get_solver)

            with sw.Client(**conf) as client:
                solvers = {s.id for s in base_client.get_solvers(software=True)}
                if solvers:
                    self.assertIn(client.get_solver().id, solvers)
                else:
                    self.assertRaises(SolverError, client.get_solver)

            with hybrid.Client(**conf) as client:
                solvers = {s.id for s in base_client.get_solvers(hybrid=True)}
                if solvers:
                    self.assertIn(client.get_solver().id, solvers)
                else:
                    self.assertRaises(SolverError, client.get_solver)


@mock.patch.multiple(Client, get_regions=get_default_regions)
class ClientConstruction(unittest.TestCase):
    """Test Client constructor and Client.from_config() factory."""

    def setUp(self):
        # prevent run-time env vars interferring with tests
        self._env = isolated_environ(empty=True).start()

    def tearDown(self):
        self._env.stop()

    def test_default(self):
        conf = {k: k for k in 'endpoint token'.split()}
        with mock.patch("dwave.cloud.client.base.load_config", lambda **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.endpoint, 'endpoint')
                self.assertEqual(client.token, 'token')
                self.assertIsInstance(client, dwave.cloud.client.Client)
                self.assertNotIsInstance(client, dwave.cloud.sw.Client)
                self.assertNotIsInstance(client, dwave.cloud.qpu.Client)
                self.assertNotIsInstance(client, dwave.cloud.hybrid.Client)

    def test_region_default(self):
        conf = {k: k for k in 'token'.split()}
        with mock.patch("dwave.cloud.client.base.load_config", lambda **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.region, constants.DEFAULT_REGION)
                self.assertEqual(client.endpoint, constants.DEFAULT_SOLVER_API_ENDPOINT)
                self.assertEqual(client.token, 'token')
                self.assertIsInstance(client, dwave.cloud.client.Client)
                self.assertNotIsInstance(client, dwave.cloud.sw.Client)
                self.assertNotIsInstance(client, dwave.cloud.qpu.Client)
                self.assertNotIsInstance(client, dwave.cloud.hybrid.Client)

    def test_region_selection_over_defaults(self):
        conf = {k: k for k in 'token'.split()}
        region_code = 'region-code'
        region_endpoint = 'region-endpoint'
        region_conf = {region_code: {"endpoint": region_endpoint}}

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with mock.patch("dwave.cloud.Client.get_regions", lambda s: region_conf):
                with dwave.cloud.Client.from_config(region=region_code) as client:
                    self.assertEqual(client.region, region_code)
                    self.assertEqual(client.endpoint, region_endpoint)

    def test_region_kwarg_overrides_endpoint_from_config(self):
        conf = {k: k for k in 'endpoint token'.split()}
        region_code = 'region-code'
        region_endpoint = 'region-endpoint'
        region_conf = {region_code: {"endpoint": region_endpoint}}

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with mock.patch("dwave.cloud.Client.get_regions", lambda s: region_conf):
                with dwave.cloud.Client.from_config(region=region_code) as client:
                    self.assertEqual(client.region, region_code)
                    self.assertEqual(client.endpoint, region_endpoint)

    def test_region_endpoint_pair_kwarg_overrides_region_endpoint_pair_from_config(self):
        conf = {k: k for k in 'region endpoint token'.split()}
        region_code = 'region-code'
        region_endpoint = 'region-endpoint'
        region_conf = {region_code: {"endpoint": region_endpoint}}

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with mock.patch("dwave.cloud.Client.get_regions", lambda s: region_conf):
                with dwave.cloud.Client.from_config(region=region_code, endpoint=region_endpoint) as client:
                    self.assertEqual(client.region, region_code)
                    self.assertEqual(client.endpoint, region_endpoint)

    def test_region_from_env_overrides_endpoint_from_config(self):
        conf = {k: k for k in 'endpoint token'.split()}
        region_code = 'region-code'
        region_endpoint = 'region-endpoint'
        region_conf = {region_code: {"endpoint": region_endpoint}}

        with isolated_environ(add={'DWAVE_API_REGION': region_code}):
            with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
                with mock.patch("dwave.cloud.Client.get_regions", lambda s: region_conf):
                    with dwave.cloud.Client.from_config() as client:
                        self.assertEqual(client.region, region_code)
                        self.assertEqual(client.endpoint, region_endpoint)

    def test_metadata_api_endpoint_from_env_accepted(self):
        metadata_api_endpoint = 'metadata-endpoint'
        with isolated_environ(add={'DWAVE_METADATA_API_ENDPOINT': metadata_api_endpoint}):
            conf = {k: k for k in 'endpoint metadata_api_endpoint token'.split()}
            with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
                with dwave.cloud.Client.from_config() as client:
                    self.assertEqual(client.metadata_api_endpoint, metadata_api_endpoint)

    def test_client_type(self):
        conf = {k: k for k in 'endpoint token'.split()}

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertIsInstance(client, dwave.cloud.client.Client)

            with dwave.cloud.client.Client.from_config() as client:
                self.assertIsInstance(client, dwave.cloud.client.Client)

            with dwave.cloud.qpu.Client.from_config() as client:
                self.assertIsInstance(client, dwave.cloud.qpu.Client)

            with dwave.cloud.sw.Client.from_config() as client:
                self.assertIsInstance(client, dwave.cloud.sw.Client)

            with dwave.cloud.hybrid.Client.from_config() as client:
                self.assertIsInstance(client, dwave.cloud.hybrid.Client)

            with dwave.cloud.qpu.Client.from_config(client='base') as client:
                self.assertIsInstance(client, dwave.cloud.client.Client)

            with dwave.cloud.Client.from_config(client='qpu') as client:
                self.assertIsInstance(client, dwave.cloud.qpu.Client)

            with dwave.cloud.Client.from_config(client='sw') as client:
                self.assertIsInstance(client, dwave.cloud.sw.Client)

            with dwave.cloud.Client.from_config(client='hybrid') as client:
                self.assertIsInstance(client, dwave.cloud.hybrid.Client)

    def test_custom_kwargs(self):
        conf = {k: k for k in 'endpoint token'.split()}
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with mock.patch("dwave.cloud.client.Client.__init__", return_value=None) as init:
                dwave.cloud.Client.from_config(custom='custom')
                init.assert_called_once_with(
                    endpoint='endpoint', token='token', custom='custom')

    def test_custom_kwargs_overrides_config(self):
        conf = {k: k for k in 'endpoint token custom'.split()}
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with mock.patch("dwave.cloud.client.Client.__init__", return_value=None) as init:
                dwave.cloud.Client.from_config(custom='new-custom')
                init.assert_called_once_with(
                    endpoint='endpoint', token='token', custom='new-custom')

    def test_solver_features_from_config(self):
        solver_def = {"qpu": True}
        conf = {k: k for k in 'endpoint token'.split()}
        conf.update(solver=json.dumps(solver_def))

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.default_solver, solver_def)

    def test_solver_name_from_config(self):
        solver_def = {"name__eq": "solver"}
        conf = {k: k for k in 'endpoint token solver'.split()}

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.default_solver, solver_def)

    def test_solver_features_kwargs_override_config(self):
        new_solver_def = {"software": True}
        conf = {k: k for k in 'endpoint token solver'.split()}

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(solver=new_solver_def) as client:
                self.assertEqual(client.default_solver, new_solver_def)

    def test_none_kwargs_do_not_override_config(self):
        """kwargs with value ``None`` should be ignored (issue #430)"""
        conf = {k: k for k in 'endpoint token'.split()}
        solver_json = '{"qpu": true}'
        conf.update(solver=solver_json)
        solver = json.loads(solver_json)

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(endpoint=None, solver=None) as client:
                self.assertEqual(client.endpoint, conf['endpoint'])
                self.assertEqual(client.default_solver, solver)

    def test_solver_name_overrides_config_features(self):
        conf = {k: k for k in 'endpoint token solver'.split()}
        conf.update(solver=json.dumps({"software": True}))

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(solver='solver') as client:
                self.assertEqual(client.default_solver, {"name__eq": "solver"})

    def test_boolean_options_parsed_from_config(self):
        conf = {'connection_close': 'off', 'permissive_ssl': 'true'}

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(token='token') as client:
                self.assertFalse(client.connection_close)
                self.assertTrue(client.permissive_ssl)

    def test_class_defaults(self):
        token = 'value'
        DEFAULTS = Client.DEFAULTS.copy()
        DEFAULTS.update(token=token)

        with mock.patch("dwave.cloud.client.base.load_config", lambda **kw: {}):
            with mock.patch.multiple("dwave.cloud.Client", DEFAULTS=DEFAULTS):
                with dwave.cloud.Client.from_config() as client:
                    self.assertEqual(client.token, token)

                # None defaults are ignored
                with dwave.cloud.Client(defaults=None) as client:
                    self.assertEqual(client.token, token)

                # explicit None kwargs do not modify defaults
                with dwave.cloud.Client(
                        endpoint=None, token=None, solver=None,
                        connection_close=None, poll_backoff_min=None) as client:

                    self.assertEqual(client.endpoint, client.DEFAULT_API_ENDPOINT)
                    self.assertEqual(client.token, token)
                    self.assertEqual(client.default_solver, {})

                    self.assertEqual(client.connection_close, DEFAULTS['connection_close'])
                    self.assertEqual(client.poll_backoff_min, DEFAULTS['poll_backoff_min'])

    def test_defaults_as_kwarg(self):
        token = 'value'
        defaults = dict(token=token)

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: {}):
            with dwave.cloud.Client.from_config(defaults=defaults) as client:
                self.assertEqual(client.token, token)

    def test_defaults_partial_update(self):
        """Some options come from DEFAULTS, some from defaults, some from config, and some from kwargs"""

        token = 'value'
        solver = {'feature': 'value'}
        request_timeout = 10

        DEFAULTS = Client.DEFAULTS.copy()
        DEFAULTS.update(token='wrong')

        defaults = dict(solver='wrong')

        conf = dict(solver=solver, request_timeout=request_timeout)

        kwargs = dict(token=token, defaults=defaults, request_timeout=None)

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with mock.patch.multiple("dwave.cloud.Client", DEFAULTS=DEFAULTS):
                with dwave.cloud.Client.from_config(**kwargs) as client:

                    # token: set on DEFAULTS, overwritten in kwargs
                    self.assertEqual(client.token, token)

                    # solver: set in defaults, overwritten in file conf
                    self.assertEqual(client.default_solver, solver)

                    # endpoint: derived from region
                    self.assertEqual(client.endpoint, constants.DEFAULT_SOLVER_API_ENDPOINT)

                    # None kwarg: used from class defaults
                    self.assertEqual(client.request_timeout, request_timeout)

    def test_headers_from_config(self):
        headers_dict = {"key-1": "value-1", "key-2": "value-2"}
        headers_str = """  key-1:value-1
            key-2: value-2
        """
        conf = dict(token='token', headers=headers_str)

        with mock.patch("dwave.cloud.client.base.load_config", lambda **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertDictEqual(client.headers, headers_dict)

    def test_headers_from_kwargs(self):
        headers_dict = {"key-1": "value-1", "key-2": "value-2"}
        headers_str = "key-2:value-2\nkey-1:value-1"
        conf = dict(token='token')

        # headers as dict
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(headers=headers_dict) as client:
                self.assertDictEqual(client.headers, headers_dict)

        # headers as str
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(headers=headers_str) as client:
                self.assertDictEqual(client.headers, headers_dict)

    def test_client_cert_from_config(self):
        crt = '/path/to/crt'
        key = '/path/to/key'

        # single file with cert+key
        client_cert = crt
        conf = dict(token='token', client_cert=crt)

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.client_cert, client_cert)

                session = client.create_session()
                self.assertEqual(session.cert, client_cert)

        # separate cert and key files
        client_cert = (crt, key)
        conf = dict(token='token', client_cert=crt, client_cert_key=key)

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.client_cert, client_cert)

                session = client.create_session()
                self.assertEqual(session.cert, client_cert)

    def test_client_cert_from_kwargs(self):
        crt = '/path/to/crt'
        key = '/path/to/key'

        def load_config(**kwargs):
            file_conf = dict(token='token')
            return merge(kwargs, file_conf, op=lambda a, b: a or b)

        # single file with cert+key
        client_cert = crt
        conf = dict(client_cert=crt)

        with mock.patch("dwave.cloud.client.base.load_config", load_config):
            with dwave.cloud.Client.from_config(**conf) as client:
                self.assertEqual(client.client_cert, client_cert)

        # separate cert and key files
        client_cert = (crt, key)
        conf = dict(client_cert=crt, client_cert_key=key)

        with mock.patch("dwave.cloud.client.base.load_config", load_config):
            with dwave.cloud.Client.from_config(**conf) as client:
                self.assertEqual(client.client_cert, client_cert)

        # client_cert as tuple (direct `requests` format)
        client_cert = (crt, key)
        conf = dict(client_cert=client_cert)

        with mock.patch("dwave.cloud.client.base.load_config", load_config):
            with dwave.cloud.Client.from_config(**conf) as client:
                self.assertEqual(client.client_cert, client_cert)

    def test_polling_params_from_config(self):
        poll_conf = {"poll_backoff_min": "0.1", "poll_backoff_max": "1"}
        conf = dict(token='token', **poll_conf)

        # polling params from config file propagated to client object
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.poll_backoff_min, 0.1)
                self.assertEqual(client.poll_backoff_max, 1.0)

        # test defaults
        conf = dict(token='token')
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.poll_backoff_min, Client.DEFAULTS['poll_backoff_min'])
                self.assertEqual(client.poll_backoff_max, Client.DEFAULTS['poll_backoff_max'])

    def test_polling_params_from_kwargs(self):
        poll_conf = {"poll_backoff_min": "0.1", "poll_backoff_max": "1"}
        conf = dict(token='token', **poll_conf)

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(poll_backoff_min=0.5) as client:
                self.assertEqual(client.poll_backoff_min, 0.5)
                self.assertEqual(client.poll_backoff_max, 1.0)

    def _verify_retry_config(self, retry, opts):
        self.assertEqual(retry.total, opts['http_retry_total'])
        self.assertEqual(retry.connect, opts['http_retry_connect'])
        self.assertEqual(retry.read, opts['http_retry_read'])
        self.assertEqual(retry.redirect, opts['http_retry_redirect'])
        self.assertEqual(retry.status, opts['http_retry_status'])
        self.assertEqual(retry.backoff_factor, opts['http_retry_backoff_factor'])
        self.assertEqual(retry.BACKOFF_MAX, opts['http_retry_backoff_max'])

    def test_http_retry_params_from_config(self):
        retry_opts = {
            "http_retry_total": 3,
            "http_retry_connect": 2,
            "http_retry_read": 2,
            "http_retry_redirect": 0,
            "http_retry_status": 2,
            "http_retry_backoff_factor": 0.5,
            "http_retry_backoff_max": 30,
        }
        retry_conf = {k: str(v) for k, v in retry_opts.items()}
        conf = dict(token='token', **retry_conf)

        # http retry params from config file propagated to client object
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                for opt, val in retry_opts.items():
                    self.assertEqual(getattr(client, opt), val,
                                     "%s doesn't match" % opt)

                # verify Retry object config
                retry = client.session.get_adapter('https://').max_retries
                self._verify_retry_config(retry, retry_opts)

        # test defaults
        conf = dict(token='token')
        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                for param in retry_conf:
                    self.assertEqual(getattr(client, param), Client.DEFAULTS[param])

    def test_http_retry_params_from_kwargs(self):
        retry_kwargs = {
            "http_retry_total": 3,
            "http_retry_connect": 2,
            "http_retry_read": None,
            "http_retry_redirect": 0,
            "http_retry_status": None,
            "http_retry_backoff_factor": 0.5,
            "http_retry_backoff_max": 30,
        }
        conf = dict(token='token')

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with dwave.cloud.Client.from_config(**retry_kwargs) as client:
                # verify client final config
                for opt, val in retry_kwargs.items():
                    self.assertEqual(getattr(client, opt), val,
                                     "%s doesn't match" % opt)

                # verify Retry object config
                retry = client.session.get_adapter('https://').max_retries
                self._verify_retry_config(retry, retry_kwargs)

    def test_positional_args(self):
        with mock.patch("dwave.cloud.client.base.load_config", lambda **kw: {}):

            # up to 3 positional args are allowed
            endpoint, token, solver = 'endpoint token solver'.split()
            with dwave.cloud.Client(endpoint, token, solver) as client:
                self.assertEqual(client.endpoint, endpoint)
                self.assertEqual(client.token, token)
                self.assertEqual(client.default_solver, {'name__eq': solver})

            # but they are deprecated
            with self.assertWarns(DeprecationWarning):
                with dwave.cloud.Client(endpoint, token):
                    pass

            # and more than 3 are not allowed
            with self.assertRaises(TypeError):
                dwave.cloud.Client(endpoint, token, solver, 'another')

            # with clashes between pargs and kwargs not allowed as well
            with self.assertRaises(TypeError):
                dwave.cloud.Client(endpoint, endpoint=endpoint)

            # even though non-conflicting combination is allowed
            with dwave.cloud.Client(endpoint, token=token) as client:
                self.assertEqual(client.endpoint, endpoint)
                self.assertEqual(client.token, token)


@mock.patch.multiple(Client, get_regions=get_default_regions)
class ClientConfigIntegration(unittest.TestCase):

    def test_custom_options(self):
        """Test custom options (request_timeout, polling_timeout, permissive_ssl)
        are propagated to Client."""

        request_timeout = 15
        polling_timeout = 180

        config_body = """
            [custom]
            token = 123
            permissive_ssl = on
            request_timeout = {}
            polling_timeout = {}
        """.format(request_timeout, polling_timeout)

        with mock.patch("dwave.cloud.config.open", iterable_mock_open(config_body)):
            with Client.from_config('config_file', profile='custom') as client:
                # check permissive_ssl and timeouts custom params passed-thru
                self.assertFalse(client.session.verify)
                self.assertEqual(client.request_timeout, request_timeout)
                self.assertEqual(client.polling_timeout, polling_timeout)

                # verify client uses those properly
                def mock_send(*args, **kwargs):
                    self.assertEqual(kwargs.get('timeout'), request_timeout)
                    response = requests.Response()
                    response.status_code = 200
                    response._content = b'{}'
                    return response

                with mock.patch("requests.adapters.HTTPAdapter.send", mock_send):
                    client.get_solvers()


class MultiRegionSupport(unittest.TestCase):

    def test_region_selection_mocked_end_to_end(self):
        """Given `metadata_api_endpoint` and `region` in a config file,
        when Client is initialized from config,
        then it is set-up with sapi endpoint retrieved from the metadata api."""

        metadata_api_endpoint = 'metadata-api'
        regions_response = [
            {"code": "a", "endpoint": "endpoint-a", "name": "A"},
            {"code": "b", "endpoint": "endpoint-b", "name": "B"}
        ]
        selected_region = regions_response[0]['code']
        selected_endpoint = regions_response[0]['endpoint']

        config_body = f"""
            [defaults]
            metadata_api_endpoint = {metadata_api_endpoint}
            region = {selected_region}
            token = token
        """

        expected_get_regions_return = {
            r['code']: {"name": r['name'], "endpoint": r['endpoint']}
            for r in regions_response
        }

        def _patch_session(resource):
            self.assertEqual(resource.session.base_url, f"{metadata_api_endpoint}/")
            session = mock.Mock()

            def get(url, **kwargs):
                response = mock.Mock(['text', 'json'])
                response.status_code = 200
                response.json.side_effect = lambda: regions_response
                return response

            session.get = get
            resource.session = session

        with mock.patch("dwave.cloud.config.open", iterable_mock_open(config_body)):
            with mock.patch.multiple(Client._fetch_available_regions._cached, cache={}):
                with mock.patch("dwave.cloud.client.base.api.Regions._patch_session", _patch_session):
                    # note: we specify config_file, otherwise reading from files
                    # might be skipped altogether if zero config files found on disk
                    # (i.e. config open mock above fails on CI)
                    with Client.from_config(config_file='config_file') as client:
                        # verify get_regions() returns data from metadata_api_endpoint
                        regions = client.get_regions()
                        self.assertEqual(regions, expected_get_regions_return)

                        # check region endpoint initialized from custom metadata_api_endpoint
                        self.assertEqual(client.metadata_api_endpoint, metadata_api_endpoint)
                        self.assertEqual(client.region, selected_region)
                        self.assertEqual(client.endpoint, selected_endpoint)

    def test_region_endpoint_fallback_when_no_metadata_api(self):
        """Given region in config, and endpoint omitted,
        when Client is initialized from config and MetadataAPI is down,
        then the client is initialized to use the default (old) endpoint and region.
        """
        # test Client._resolve_region_endpoint

        conf = dict(token='token', metadata_api_endpoint='invalid')
        region = Client.DEFAULT_API_REGION

        with mock.patch("dwave.cloud.client.base.load_config", lambda **kw: conf):
            with dwave.cloud.Client.from_config(region=region) as client:
                self.assertEqual(client.region, Client.DEFAULT_API_REGION)
                self.assertEqual(client.endpoint, Client.DEFAULT_API_ENDPOINT)

    @mock.patch.multiple(Client, get_regions=get_default_regions)
    def test_region_endpoint_fallback_when_region_unknown(self):
        """Given invalid region in config, and endpoint omitted,
        when Client is initialized from config,
        then ValueError is raised.
        """
        # test Client._resolve_region_endpoint

        conf = dict(token='token')
        region = 'invalid-region-code'

        with mock.patch("dwave.cloud.config.load_profile_from_files", lambda *pa, **kw: conf):
            with self.assertRaises(ValueError):
                Client.from_config(region=region)

    @mock.patch.multiple(Client, get_regions=get_default_regions)
    def test_region_endpoint_null_case(self):
        """Given region as None, and endpoint as None,
        when Client is initialized from config,
        then the client is initialized to use the default (old) endpoint and region.
        """
        # test Client._resolve_region_endpoint

        conf = dict(token='token', region='')

        with mock.patch("dwave.cloud.client.base.load_config", lambda **kw: conf):
            with dwave.cloud.Client.from_config() as client:
                self.assertEqual(client.region, Client.DEFAULT_API_REGION)
                self.assertEqual(client.endpoint, Client.DEFAULT_API_ENDPOINT)

    @unittest.skipUnless(config, "No live server configuration available.")
    @mock.patch.multiple(Client._fetch_available_regions._cached, cache={})
    def test_region_selection_live_end_to_end(self):
        with Client(**config) as client:
            self.assertEqual(client.region, constants.DEFAULT_REGION)
            self.assertIn(constants.DEFAULT_REGION, client.endpoint)
            self.assertGreater(len(client.get_regions()), 0)


class FeatureBasedSolverSelection(unittest.TestCase):
    """Test Client.get_solvers(**filters)."""

    def setUp(self):
        # mock solvers
        self.qpu1 = StructuredSolver(client=None, data={
            "properties": {
                "supported_problem_types": ["qubo", "ising"],
                "qubits": [0, 1, 2],
                "couplers": [[0, 1], [0, 2], [1, 2]],
                "num_qubits": 3,
                "num_reads_range": [0, 100],
                "parameters": {
                    "num_reads": "Number of samples to return.",
                    "postprocess": "either 'sampling' or 'optimization'"
                },
                "topology": {
                    "type": "chimera",
                    "shape": [16, 16, 4]
                },
                "category": "qpu",
                "tags": ["lower_noise"]
            },
            "id": "qpu1",
            "description": "QPU Chimera solver",
            "status": "online",
            "avg_load": 0.1
        })
        self.qpu2 = StructuredSolver(client=None, data={
            "properties": {
                "supported_problem_types": ["qubo", "ising"],
                "qubits": [0, 1, 2, 3, 4],
                "couplers": [[0, 1], [0, 2], [1, 2], [2, 3], [3, 4]],
                "num_qubits": 5,
                "num_reads_range": [0, 200],
                "parameters": {
                    "num_reads": "Number of samples to return.",
                    "flux_biases": "Supported ...",
                    "anneal_schedule": "Supported ..."
                },
                "topology": {
                    "type": "pegasus",
                    "shape": [6, 6, 12]
                },
                "category": "qpu",
                "vfyc": True
            },
            "id": "qpu2",
            "description": "QPU Pegasus solver",
            "avg_load": 0.2
        })
        self.software = StructuredSolver(client=None, data={
            "properties": {
                "supported_problem_types": ["qubo", "ising"],
                "qubits": [0, 1],
                "couplers": [[0, 1]],
                "num_qubits": 7,
                "num_reads_range": [0, 1000],
                "parameters": {"num_reads": "Number of samples to return."},
                "vfyc": False,
                "topology": {
                    "type": "chimera",
                    "shape": [4, 4, 4]
                },
                "category": "software",
                # the following are only present in this solver
                "some_set": [1, 2],
                "some_range": [1, 2],
                "some_string": "x",
                "tags": ["tag"]
            },
            "id": "sw_solver1",
            "description": "Software solver",
            "avg_load": 0.7
        })
        self.hybrid = UnstructuredSolver(client=None, data={
            "properties": {
                "supported_problem_types": ["bqm"],
                "maximum_number_of_variables": 10000,
                "maximum_time_limit_hrs": 24.0,
                "minimum_time_limit": [[1, 3.0], [1024, 3.0], [4096, 10.0], [10000, 40.0]],
                "quota_conversion_rate": 20,
                "parameters": {
                    "time_limit": ""
                },
                "category": "hybrid",
            },
            "id": "hybrid_v1",
            "description": "Hybrid solver"
        })

        self.qpu_solvers = [self.qpu1, self.qpu2]
        self.software_solvers = [self.software]
        self.hybrid_solvers = [self.hybrid]

        self.structured_solvers = self.qpu_solvers + self.software_solvers
        self.unstructured_solvers = self.hybrid_solvers

        self.solvers = self.structured_solvers + self.unstructured_solvers

        # mock client
        self.client = Client(endpoint='endpoint', token='token')
        self.client._fetch_solvers = lambda **kw: self.solvers

    def shutDown(self):
        self.client.close()

    def assertSolvers(self, container, members):
        self.assertEqual(set(container), set(members))

    def test_default(self):
        self.assertSolvers(self.client.get_solvers(), self.solvers)

    def test_online(self):
        self.assertSolvers(self.client.get_solvers(online=True), self.solvers)
        self.assertSolvers(self.client.get_solvers(online=False), [])

    def test_derived_category_properties(self):
        self.assertSolvers(self.client.get_solvers(qpu=True), self.qpu_solvers)
        self.assertSolvers(self.client.get_solvers(qpu=False), self.software_solvers + self.hybrid_solvers)
        self.assertSolvers(self.client.get_solvers(software=True), self.software_solvers)
        self.assertSolvers(self.client.get_solvers(software=False), self.qpu_solvers + self.hybrid_solvers)
        self.assertSolvers(self.client.get_solvers(hybrid=True), self.hybrid_solvers)
        self.assertSolvers(self.client.get_solvers(hybrid=False), self.qpu_solvers + self.software_solvers)

    # Test fallback for legacy solvers without the `category` property
    # TODO: remove when all production solvers are updated
    def test_derived_category_properties_without_category(self):
        "Category-based filtering works without explicit `category` property."

        @contextmanager
        def multi_solver_properties_patch(solvers, update):
            """Update properties for all `solvers` at once."""
            patchers = [mock.patch.dict(s.properties, update) for s in solvers]
            try:
                yield (p.start() for p in patchers)
            finally:
                return (p.stop() for p in patchers)

        with mock.patch.object(self.software, 'id', 'c4-sw_solver3'):
            # patch categories and re-run the category-based filtering test
            with multi_solver_properties_patch(self.solvers, {'category': ''}):
                self.test_derived_category_properties()

            # verify patching
            with multi_solver_properties_patch(self.solvers, {'category': 'x'}):
                with self.assertRaises(AssertionError):
                    self.test_derived_category_properties()

    def test_category(self):
        self.assertSolvers(self.client.get_solvers(category='qpu'), self.qpu_solvers)
        self.assertSolvers(self.client.get_solvers(category='software'), self.software_solvers)
        self.assertSolvers(self.client.get_solvers(category='hybrid'), self.hybrid_solvers)

    def test_name(self):
        self.assertSolvers(self.client.get_solvers(name='qpu1'), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(name='qpu2'), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(name='solver'), [])
        self.assertSolvers(self.client.get_solvers(name='olver1'), [])
        self.assertSolvers(self.client.get_solvers(name__regex='.*1'), [self.qpu1, self.software, self.hybrid])
        self.assertSolvers(self.client.get_solvers(name__regex='.*_v1'), [self.hybrid])
        self.assertSolvers(self.client.get_solvers(name__regex='.*[12].*'), self.solvers)
        self.assertSolvers(self.client.get_solvers(name__regex='qpu[12]'), [self.qpu1, self.qpu2])
        self.assertSolvers(self.client.get_solvers(name__regex='^qpu(1|2)$'), [self.qpu1, self.qpu2])

    def test_num_qubits(self):
        self.assertSolvers(self.client.get_solvers(num_qubits=5), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(num_active_qubits=2), [self.software])
        self.assertSolvers(self.client.get_solvers(num_active_qubits__in=[2, 3]), [self.qpu1, self.software])

    def test_lower_noise_derived_property(self):
        self.assertSolvers(self.client.get_solvers(lower_noise=True), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(lower_noise=False), [self.qpu2, self.software])

    def test_parameter_availability_check(self):
        self.assertSolvers(self.client.get_solvers(postprocess__available=True), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(postprocess=True), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(parameters__contains='flux_biases'), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(parameters__contains='num_reads'), self.structured_solvers)

    def test_property_availability_check(self):
        self.assertSolvers(self.client.get_solvers(vfyc__available=True), [self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(vfyc__eq=True), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(vfyc=True), [self.qpu2])

        # inverse of vfyc=True
        self.assertSolvers(self.client.get_solvers(vfyc__in=[False, None]), [self.qpu1, self.software, self.hybrid])

        # vfyc unavailable or unadvertized
        self.assertSolvers(self.client.get_solvers(vfyc__available=False), [self.qpu1, self.hybrid])
        self.assertSolvers(self.client.get_solvers(vfyc__eq=False), [self.software])
        self.assertSolvers(self.client.get_solvers(vfyc=False), [self.software])

        # non-existing params/props have value of None
        self.assertSolvers(self.client.get_solvers(vfyc__eq=None), [self.qpu1, self.hybrid])
        self.assertSolvers(self.client.get_solvers(vfyc=None), [self.qpu1, self.hybrid])

    def test_availability_combo(self):
        self.assertSolvers(self.client.get_solvers(vfyc=False, flux_biases=True), [])
        self.assertSolvers(self.client.get_solvers(vfyc=True, flux_biases=True), [self.qpu2])

    def test_relational_ops(self):
        self.assertSolvers(self.client.get_solvers(num_qubits=3), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(num_qubits__eq=3), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(num_qubits=4), [])
        self.assertSolvers(self.client.get_solvers(num_qubits=5), [self.qpu2])

        self.assertSolvers(self.client.get_solvers(num_qubits__gte=2), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_qubits__gte=5), [self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(num_qubits__gt=5), [self.software])

        self.assertSolvers(self.client.get_solvers(num_qubits__lte=8), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_qubits__lte=7), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_qubits__lt=7), [self.qpu1, self.qpu2])

        # skip solver if LHS value not defined (None)
        self.assertSolvers(self.client.get_solvers(avg_load__gt=0), [self.qpu1, self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(avg_load__gte=0), [self.qpu1, self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(avg_load__lt=1), [self.qpu1, self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(avg_load__lte=1), [self.qpu1, self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(avg_load=0.7), [self.software])
        self.assertSolvers(self.client.get_solvers(avg_load__eq=0.7), [self.software])
        self.assertSolvers(self.client.get_solvers(avg_load=None), [self.hybrid])
        self.assertSolvers(self.client.get_solvers(avg_load__eq=None), [self.hybrid])

    def test_range_ops(self):
        # value within range
        self.assertSolvers(self.client.get_solvers(num_qubits__within=[3, 7]), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_qubits__within=[3, 5]), [self.qpu1, self.qpu2])
        self.assertSolvers(self.client.get_solvers(num_qubits__within=[2, 4]), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(num_qubits__within=(2, 4)), [self.qpu1])
        self.assertSolvers(self.client.get_solvers(num_qubits__within=(4, 2)), [self.qpu1])

        # range within (covered by) range
        self.assertSolvers(self.client.get_solvers(num_reads_range__within=(0, 500)), [self.qpu1, self.qpu2])
        self.assertSolvers(self.client.get_solvers(num_reads_range__within=(1, 500)), [])

        # invalid LHS
        self.assertSolvers(self.client.get_solvers(some_range__within=[0, 2]), [self.software])

        # range covering a value (value included in range)
        self.assertSolvers(self.client.get_solvers(num_reads_range__covers=0), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_reads_range__covers=150), [self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(num_reads_range__covers=550), [self.software])
        self.assertSolvers(self.client.get_solvers(num_reads_range__covers=1000), [self.software])
        self.assertSolvers(self.client.get_solvers(num_reads_range__covers=1001), [])

        # range covering a range
        self.assertSolvers(self.client.get_solvers(num_reads_range__covers=(10, 90)), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_reads_range__covers=(110, 200)), [self.qpu2, self.software])

        # invalid LHS
        self.assertSolvers(self.client.get_solvers(some_range__covers=1.5), [self.software])

    def test_membership_ops(self):
        # property contains
        self.assertSolvers(self.client.get_solvers(supported_problem_types__contains="qubo"), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(supported_problem_types__contains="undef"), [])
        self.assertSolvers(self.client.get_solvers(supported_problem_types__contains="bqm"), self.unstructured_solvers)
        self.assertSolvers(self.client.get_solvers(couplers__contains=[0, 1]), [self.qpu1, self.qpu2, self.software])
        self.assertSolvers(self.client.get_solvers(couplers__contains=[0, 2]), [self.qpu1, self.qpu2])

        # property in
        self.assertSolvers(self.client.get_solvers(num_qubits__in=[3, 5]), [self.qpu1, self.qpu2])
        self.assertSolvers(self.client.get_solvers(num_qubits__in=[7]), [self.software])
        self.assertSolvers(self.client.get_solvers(num_qubits__in=[]), [])

        # invalid LHS
        self.assertSolvers(self.client.get_solvers(some_set__contains=1), [self.software])
        self.assertSolvers(self.client.get_solvers(avg_load__in=[None]), [self.hybrid])
        self.assertSolvers(self.client.get_solvers(avg_load__in=[None, 0.1, 0.2, 0.7]), self.solvers)

    def test_set_ops(self):
        # property issubset
        self.assertSolvers(self.client.get_solvers(supported_problem_types__issubset=("qubo", "ising", "bqm", "other")), self.solvers)
        self.assertSolvers(self.client.get_solvers(supported_problem_types__issubset=["qubo", "ising", "bqm"]), self.solvers)
        self.assertSolvers(self.client.get_solvers(supported_problem_types__issubset=["bqm", "ising", "qubo"]), self.solvers)
        self.assertSolvers(self.client.get_solvers(supported_problem_types__issubset={"ising", "qubo", "bqm"}), self.solvers)
        self.assertSolvers(self.client.get_solvers(supported_problem_types__issubset=("unicorn", "ising", "other")), [])

        self.assertSolvers(self.client.get_solvers(supported_problem_types__issubset={"ising", "qubo"}), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(supported_problem_types__issubset={"bqm", "other"}), self.unstructured_solvers)

        # property issuperset
        self.assertSolvers(self.client.get_solvers(qubits__issuperset={0, 1}), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(qubits__issuperset={1, 2}), [self.qpu1, self.qpu2])

        # unhashable types
        self.assertSolvers(self.client.get_solvers(couplers__issuperset=[[0, 1]]), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(couplers__issuperset={(0, 1)}), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(couplers__issuperset={(0, 1), (0, 2)}), [self.qpu1, self.qpu2])
        self.assertSolvers(self.client.get_solvers(couplers__issuperset={(0, 1), (0, 2), (2, 3)}), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(couplers__issuperset={(0, 1), (0, 2), (2, 3), (0, 5)}), [])

        # invalid LHS
        self.assertSolvers(self.client.get_solvers(some_set__issubset={0, 1, 2}), [self.software])
        self.assertSolvers(self.client.get_solvers(some_set__issuperset={1}), [self.software])

    def test_regex(self):
        self.assertSolvers(self.client.get_solvers(num_reads__regex='.*number.*'), [])
        self.assertSolvers(self.client.get_solvers(num_reads__regex='.*Number.*'), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_reads__regex='Number.*'), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(num_reads__regex='Number'), [])

        # invalid LHS
        self.assertSolvers(self.client.get_solvers(some_string__regex='x'), [self.software])

    def test_range_boolean_combo(self):
        self.assertSolvers(self.client.get_solvers(num_qubits=3, vfyc=True), [])
        self.assertSolvers(self.client.get_solvers(num_qubits__gte=3, vfyc=True), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(num_qubits__lte=4, vfyc=True), [])
        self.assertSolvers(self.client.get_solvers(num_qubits__within=(3, 6), flux_biases=True), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(num_qubits=5, flux_biases=True), [self.qpu2])

    def test_nested_properties_leaf_lookup(self):
        self.assertSolvers(self.client.get_solvers(topology__type="chimera"), [self.qpu1, self.software])
        self.assertSolvers(self.client.get_solvers(topology__type="pegasus"), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(topology__type__eq="pegasus"), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(topology__shape=[6,6,12]), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(topology__type="chimera", topology__shape__contains=16), [self.qpu1])

    def test_nested_properties_intermediate_key_lookup(self):
        self.assertSolvers(self.client.get_solvers(topology__contains="shape"), self.structured_solvers)
        self.assertSolvers(self.client.get_solvers(topology={"type": "pegasus", "shape": [6, 6, 12]}), [self.qpu2])

    def test_anneal_schedule(self):
        self.assertSolvers(self.client.get_solvers(anneal_schedule__available=True), [self.qpu2])
        self.assertSolvers(self.client.get_solvers(anneal_schedule=True), [self.qpu2])

    def test_order_by_edgecases(self):
        # default: sort by avg_load
        self.assertEqual(self.client.get_solvers(), [self.qpu1, self.qpu2, self.software, self.hybrid])

        # explicit no sort
        self.assertEqual(self.client.get_solvers(order_by=None), self.solvers)
        self.assertEqual(self.client.get_solvers(order_by=''), self.solvers)
        self.assertEqual(self.client.get_solvers(order_by=False), self.solvers)

        # reverse without sorting
        self.assertEqual(self.client.get_solvers(order_by='-'), list(reversed(self.solvers)))

        # invalid type of `order_by`
        with self.assertRaises(TypeError):
            self.client.get_solvers(order_by=1)
        with self.assertRaises(TypeError):
            self.client.get_solvers(order_by=list)

    def test_order_by_respects_default_solver(self):
        """order_by used in isolation should not affect default_solver filters (issue #401)"""

        with Client(endpoint='endpoint', token='token', solver=dict(name='qpu2')) as client:
            # mock the network call to fetch all solvers
            client._fetch_solvers = lambda **kw: self.solvers

            # the default solver was set on client init
            self.assertEqual(client.get_solver(), self.qpu2)

            # the default solver should not change when we add order_by
            self.assertEqual(client.get_solver(order_by='id'), self.qpu2)

        with Client(endpoint='endpoint', token='token', solver=dict(category='qpu')) as client:
            # mock the network call to fetch all solvers
            client._fetch_solvers = lambda **kw: self.solvers

            # test default order_by is avg_load
            self.assertEqual(client.get_solver(), self.qpu1)

            # but we can change it, without affecting solver filters
            self.assertEqual(client.get_solver(order_by='-avg_load'), self.qpu2)

    def test_order_by_in_default_solver(self):
        """order_by can be specified as part of default_solver filters (issue #407)"""

        with Client(endpoint='endpoint', token='token', solver=dict(order_by='id')) as client:
            # mock the network call to fetch all solvers
            client._fetch_solvers = lambda **kw: self.solvers

            # the default solver was set on client init
            self.assertEqual(client.get_solver(), self.hybrid)

            # the default solver can be overridden
            self.assertEqual(client.get_solver(order_by='-id'), self.software)

        with Client(endpoint='endpoint', token='token', solver=dict(qpu=True, order_by='-num_active_qubits')) as client:
            # mock the network call to fetch all solvers
            client._fetch_solvers = lambda **kw: self.solvers

            # the default solver was set on client init
            self.assertEqual(client.get_solver(), self.qpu2)

            # adding order_by doesn't change other default solver features
            self.assertEqual(client.get_solver(order_by='num_active_qubits'), self.qpu1)

    def test_order_by_string(self):
        # sort by Solver inferred properties
        self.assertEqual(self.client.get_solvers(order_by='id'), [self.hybrid, self.qpu1, self.qpu2, self.software])
        self.assertEqual(self.client.get_solvers(order_by='qpu'), [self.software, self.hybrid, self.qpu1, self.qpu2])
        self.assertEqual(self.client.get_solvers(order_by='num_qubits'), self.solvers)
        self.assertEqual(self.client.get_solvers(order_by='num_active_qubits'), [self.software, self.qpu1, self.qpu2, self.hybrid])

        # sort by solver property
        self.assertEqual(self.client.get_solvers(order_by='properties.num_qubits'), self.solvers)

        # sort (and reverse sort) by upper bound of a range property
        self.assertEqual(self.client.get_solvers(order_by='properties.num_reads_range[1]'), self.solvers)
        self.assertEqual(self.client.get_solvers(order_by='-properties.num_reads_range[1]'), [self.hybrid, self.software, self.qpu2, self.qpu1])

        # check solvers with None values for key end up last
        self.assertEqual(self.client.get_solvers(order_by='properties.vfyc'), [self.software, self.qpu2, self.qpu1, self.hybrid])
        self.assertEqual(self.client.get_solvers(order_by='-properties.vfyc'), [self.hybrid, self.qpu1, self.qpu2, self.software])

        # check invalid keys don't fail, and effectively don't sort the list
        self.assertEqual(self.client.get_solvers(order_by='non_existing_key'), self.solvers)
        self.assertEqual(self.client.get_solvers(order_by='-non_existing_key'), list(reversed(self.solvers)))

    def test_order_by_callable(self):
        # sort by Solver inferred properties
        self.assertEqual(self.client.get_solvers(order_by=lambda solver: solver.id), [self.hybrid, self.qpu1, self.qpu2, self.software])
        self.assertEqual(self.client.get_solvers(order_by=lambda solver: solver.avg_load), [self.qpu1, self.qpu2, self.software, self.hybrid])

        # sort by solver property
        self.assertEqual(self.client.get_solvers(order_by=lambda solver: solver.properties.get('num_qubits')), self.solvers)

        # sort None`s last (here: False, True, None)
        self.assertEqual(self.client.get_solvers(order_by=lambda solver: solver.properties.get('vfyc')), [self.software, self.qpu2, self.qpu1, self.hybrid])

        # test no sort
        self.assertEqual(self.client.get_solvers(order_by=lambda solver: None), self.solvers)


if __name__ == '__main__':
    unittest.main()



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`solver` is a resource for solving problems.

Solvers are responsible for:

    - Encoding submitted problems
    - Checking submitted parameters
    - Decoding answers
    - Adding problems to a client's submission queue

You can list all solvers available to a :class:`~dwave.cloud.client.Client` with its
:func:`~dwave.cloud.client.Client.get_solvers` method and select and return one with its
:func:`~dwave.cloud.client.Client.get_solver` method.

"""

import copy
import json
import logging
import typing
from collections.abc import Mapping

from dwave.cloud.exceptions import (
    InvalidAPIResponseError, SolverPropertyMissingError,
    UnsupportedSolverError, ProblemStructureError)
from dwave.cloud.coders import (
    encode_problem_as_qp, encode_problem_as_ref,
    decode_qp_numpy, decode_qp, decode_bq, bqm_as_file)
from dwave.cloud.utils import reformat_qubo_as_ising, NumpyEncoder
from dwave.cloud.computation import Future
from dwave.cloud.concurrency import Present
from dwave.cloud.events import dispatches_events

# Use numpy if available for fast encoding/decoding
try:
    import numpy as np
    _numpy = True
except ImportError:
    _numpy = False

try:
    import dimod
except ImportError:
    dimod = None

__all__ = [
    'BaseSolver', 'StructuredSolver',
    'BaseUnstructuredSolver', 'UnstructuredSolver',
    'Solver', 'BQMSolver', 'CQMSolver', 'DQMSolver',
]

logger = logging.getLogger(__name__)


if typing.TYPE_CHECKING:
    # do a bit of fiddling
    try:
        _Type = typing.Literal['qubo', 'ising']
    except ImportError:  # Python < 3.8
        _Type = str

    try:
        _Vartype = typing.Union[_Type, dimod.typing.VartypeLike]
    except AttributeError:  # dimod not installed or too old
        _Vartype = _Type


class BaseSolver(object):
    """Base class for a general D-Wave solver.

    This class provides :term:`Ising`, :term:`QUBO` and :term:`BQM` sampling
    methods and encapsulates the solver description returned from the D-Wave
    cloud API.

    Args:
        client (:class:`Client`):
            Client that manages access to this solver.

        data (`dict`):
            Data from the server describing this solver.

    Examples:
        This example creates a client using the local system's default D-Wave Cloud
        Client configuration file and checks the identity of its default solver.

        >>> from dwave.cloud import Client
        >>> with Client.from_config() as client:
        ...     solver = client.get_solver()
        ...     solver.id       # doctest: +SKIP
        'EXAMPLE_2000Q_SYSTEM'
    """

    # Classes of problems the remote solver has to support (at least one of these)
    # in order for `Solver` to be able to abstract, or use, that solver
    _handled_problem_types = {}
    _handled_encoding_formats = {}

    def __init__(self, client, data):
        # client handles async api requests (via local thread pool)
        self.client = client

        # data for each solver includes at least: id, description, properties,
        # status and avg_load
        self.data = data

        # Each solver has an ID field
        try:
            self.id = data['id']
        except KeyError:
            raise InvalidAPIResponseError("Missing solver property: 'id'")

        # Properties of this solver the server presents: dict
        try:
            self.properties = data['properties']
        except KeyError:
            raise SolverPropertyMissingError("Missing solver property: 'properties'")

        # The set of extra parameters this solver will accept in sample_ising or sample_qubo: dict
        self.parameters = self.properties.get('parameters', {})

        # Ensure this remote solver supports at least one of the problem types we know how to handle
        try:
            self.supported_problem_types = set(self.properties['supported_problem_types'])
        except KeyError:
            raise SolverPropertyMissingError(
                "Missing solver property: 'properties.supported_problem_types'")

        if self.supported_problem_types.isdisjoint(self._handled_problem_types):
            raise UnsupportedSolverError(
                ("Remote solver {id!r} supports {supports} problems, "
                 "but {cls!r} class of solvers handles only {handled}").format(
                    id=self.id,
                    supports=list(self.supported_problem_types),
                    cls=type(self).__name__,
                    handled=list(self._handled_problem_types)))

        # When True the solution data will be returned as numpy matrices: False
        # TODO: deprecate
        self.return_matrix = False

        # Derived solver properties (not present in solver data properties dict)
        self.derived_properties = {
            'qpu', 'hybrid', 'software', 'online', 'avg_load', 'name'
        }

    def __repr__(self):
        return "{}(id={!r})".format(type(self).__name__, self.id)

    def _retrieve_problem(self, id_):
        """Resume polling for a problem previously submitted.

        Args:
            id_: Identification of the query.

        Returns:
            :class:`~dwave.cloud.computation.Future`
        """
        future = Future(self, id_, self.return_matrix)
        self.client._poll(future)
        return future

    def check_problem(self, *args, **kwargs):
        return True

    def _decode_qp(self, msg):
        if _numpy:
            return decode_qp_numpy(msg, return_matrix=self.return_matrix)
        else:
            return decode_qp(msg)

    def decode_response(self, msg):
        if msg['type'] not in self._handled_problem_types:
            raise ValueError('Unknown problem type received.')

        fmt = msg.get('answer', {}).get('format')
        if fmt not in self._handled_encoding_formats:
            raise ValueError('Unhandled answer encoding format received.')

        if fmt == 'qp':
            return self._decode_qp(msg)
        elif fmt == 'bq':
            return decode_bq(msg)
        else:
            raise ValueError("Don't know how to decode %r answer format" % fmt)

    # Sampling methods
    def sample_ising(self, linear, quadratic, **params):
        raise NotImplementedError

    def sample_qubo(self, qubo, **params):
        raise NotImplementedError

    def sample_bqm(self, bqm, **params):
        raise NotImplementedError

    def upload_bqm(self, bqm):
        raise NotImplementedError

    # Derived properties

    @property
    def name(self):
        """Solver name/ID."""
        return self.id

    @property
    def online(self):
        "Is this solver online (or offline)?"
        return self.data.get('status', 'online').lower() == 'online'

    @property
    def avg_load(self):
        "Solver's average load, at the time of description fetch."
        return self.data.get('avg_load')

    @property
    def qpu(self):
        "Is this a QPU-based solver?"
        category = self.properties.get('category', '').lower()
        if category:
            return category == 'qpu'
        else:
            # fallback for legacy solvers without the `category` property
            # TODO: remove when all production solvers are updated
            return not (self.software or self.hybrid)

    @property
    def software(self):
        "Is this a software-based solver?"
        category = self.properties.get('category', '').lower()
        if category:
            return category == 'software'
        else:
            # fallback for legacy solvers without the `category` property
            # TODO: remove when all production solvers are updated
            return self.id.startswith('c4-sw_')

    @property
    def hybrid(self):
        "Is this a hybrid quantum-classical solver?"
        category = self.properties.get('category', '').lower()
        if category:
            return category == 'hybrid'
        else:
            # fallback for legacy solvers without the `category` property
            # TODO: remove when all production solvers are updated
            return self.id.startswith('hybrid')


class BaseUnstructuredSolver(BaseSolver):
    """Base class for D-Wave unstructured solvers.

    This class provides :term:`Ising`, :term:`QUBO` and :term:`BQM` sampling
    methods and encapsulates the solver description returned from the D-Wave
    cloud API.

    Args:
        client (:class:`~dwave.cloud.client.Client`):
            Client that manages access to this solver.

        data (`dict`):
            Data from the server describing this solver.

    Note:
        Events are not yet dispatched from unstructured solvers.
    """

    def sample_ising(self, linear, quadratic, offset=0, label=None, **params):
        """Sample from the specified :term:`Ising` model.

        Args:
            linear (dict/list):
                Linear biases of the Ising problem. If a dict, should be of the
                form `{v: bias, ...}` where v is a spin-valued variable and `bias`
                is its associated bias. If a list, it is treated as a list of
                biases where the indices are the variable labels.

            quadratic (dict[(variable, variable), bias]):
                Quadratic terms of the model (J), stored in a dict. With keys
                that are 2-tuples of variables and values are quadratic biases
                associated with the pair of variables (the interaction).

            offset (optional, default=0):
                Constant offset applied to the model.

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Note:
            To use this method, dimod package has to be installed.
        """
        try:
            import dimod
        except ImportError: # pragma: no cover
            raise RuntimeError("Can't use unstructured solver without dimod. "
                               "Re-install the library with bqm/cqm/dqm support.")

        bqm = dimod.BinaryQuadraticModel.from_ising(linear, quadratic, offset)
        return self.sample_bqm(bqm, label=label, **params)

    def sample_qubo(self, qubo, offset=0, label=None, **params):
        """Sample from the specified :term:`QUBO`.

        Args:
            qubo (dict):
                Coefficients of a quadratic unconstrained binary optimization
                (QUBO) problem. Should be a dict of the form `{(u, v): bias, ...}`
                where `u`, `v`, are binary-valued variables and `bias` is their
                associated coefficient.

            offset (optional, default=0):
                Constant offset applied to the model.

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Note:
            To use this method, dimod package has to be installed.
        """
        try:
            import dimod
        except ImportError: # pragma: no cover
            raise RuntimeError("Can't use unstructured solver without dimod. "
                               "Re-install the library with bqm/cqm/dqm support.")

        bqm = dimod.BinaryQuadraticModel.from_qubo(qubo, offset)
        return self.sample_bqm(bqm, label=label, **params)

    def _encode_problem_for_upload(self, problem):
        """Encode problem for upload to solver.

        Args:
            problem (dimod-model-like):
                Problem of type `._handled_problem_types`.

        Returns:
            file-like:
                Binary stream ready for reading and seeking.
        """
        raise NotImplementedError

    def upload_problem(self, problem):
        """Encode and upload the problem.

        Args:
            problem (dimod-model-like):
                Quadratic model handled by the solver, for example
                :class:`dimod.BQM`, :class:`dimod.CQM` or :class:`dimod.DQM`.

        Returns:
            :class:`concurrent.futures.Future`[str]:
                Problem ID in a Future. Problem ID can be used to submit
                problems by reference.
        """
        data = self._encode_problem_for_upload(problem)
        return self.client.upload_problem_encoded(data)

    def _encode_problem_for_submission(self, problem, problem_type, params,
                                       label=None):
        """Encode `problem` for submitting in `ref` format. Upload the
        problem if it's not already uploaded.

        Args:
            problem (dimod-model-like/str):
                A quadratic model, or a reference to one (Problem ID).

            problem_type (str):
                Problem type, one of the handled problem types by the solver.

            params (dict):
                Parameters for the sampling method, solver-specific.

            label (str, optional):
                Problem label.

        Returns:
            str:
                JSON-encoded problem submit body
        """

        if isinstance(problem, str):
            problem_id = problem
        else:
            logger.debug("To encode the problem for submit in the 'ref' format, "
                         "we need to upload it first.")
            problem_id = self.upload_problem(problem).result()

        body = {
            'solver': self.id,
            'data': encode_problem_as_ref(problem_id),
            'type': problem_type,
            'params': params
        }
        if label is not None:
            body['label'] = label
        body_data = json.dumps(body, cls=NumpyEncoder)
        logger.trace("Sampling request encoded as: %s", body_data)

        return body_data

    @dispatches_events('sample')
    def sample_problem(self, problem, problem_type=None, label=None, **params):
        """Sample from the specified problem.

        Args:
            problem (dimod-model-like/str):
                A quadratic model (e.g. :class:`dimod.BQM`/:class:`dimod.CQM`/:class:`dimod.DQM`),
                or a reference to one (Problem ID returned by
                :meth:`.upload_problem` method).

            problem_type (str, optional):
                Problem type, one of the handled problem types by the solver.
                If not specified, the first handled problem type is used.

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`
        """

        # infer problem_type; for now just the the first handled (always just one)
        if problem_type is None:
            problem_type = next(iter(self._handled_problem_types))

        # encode the request (body as future)
        body = self.client._encode_problem_executor.submit(
            self._encode_problem_for_submission,
            problem=problem, problem_type=problem_type, params=params,
            label=label)

        # computation future holds a reference to the remote job
        computation = Future(
            solver=self, id_=None, return_matrix=self.return_matrix)

        logger.debug("Submitting new problem to: %s", self.id)
        self.client._submit(body, computation)

        return computation


class BQMSolver(BaseUnstructuredSolver):
    """Class for D-Wave unstructured binary quadratic model solvers.

    This class provides :term:`Ising`, :term:`QUBO` and :term:`BQM` sampling
    methods and encapsulates the solver description returned from the D-Wave
    cloud API.

    Args:
        client (:class:`~dwave.cloud.client.Client`):
            Client that manages access to this solver.

        data (`dict`):
            Data from the server describing this solver.

    Note:
        Events are not yet dispatched from unstructured solvers.
    """

    _handled_problem_types = {"bqm"}
    _handled_encoding_formats = {"bq"}

    def _encode_problem_for_upload(self, bqm):
        try:
            data = bqm_as_file(bqm)
        except Exception as e:
            logger.debug("BQM conversion to file failed with %r, "
                         "assuming data already encoded.", e)
            # assume `bqm` given as file, ready for upload
            data = bqm

        return data

    def sample_bqm(self, bqm, label=None, **params):
        """Sample from the specified :term:`BQM`.

        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`/str):
                A binary quadratic model, or a reference to one
                (Problem ID returned by `.upload_bqm` method).

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Note:
            To use this method, dimod package has to be installed.
        """
        return self.sample_problem(bqm, label=label, **params)

    def upload_bqm(self, bqm):
        """Upload the specified :term:`BQM` to SAPI, returning a Problem ID
        that can be used to submit the BQM to this solver (i.e. call the
        `.sample_bqm` method).

        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`/bytes-like/file-like):
                A binary quadratic model given either as an in-memory
                :class:`~dimod.BinaryQuadraticModel` object, or as raw data
                (encoded serialized model) in either a file-like or a bytes-like
                object.

        Returns:
            :class:`concurrent.futures.Future`[str]:
                Problem ID in a Future. Problem ID can be used to submit
                problems by reference.

        Note:
            To use this method, dimod package has to be installed.
        """
        return self.upload_problem(bqm)


class DQMSolver(BaseUnstructuredSolver):
    """Class for D-Wave unstructured discrete quadratic model solvers.

    This class provides a :term:`DQM` sampling
    method and encapsulates the solver description returned from the D-Wave
    cloud API.

    Args:
        client (:class:`~dwave.cloud.client.Client`):
            Client that manages access to this solver.

        data (`dict`):
            Data from the server describing this solver.

    Note:
        Events are not yet dispatched from unstructured solvers.
    """

    _handled_problem_types = {"dqm"}
    _handled_encoding_formats = {"bq"}

    def _encode_problem_for_upload(self, dqm):
        try:
            data = dqm.to_file()
        except Exception as e:
            logger.debug("DQM conversion to file failed with %r, "
                         "assuming data already encoded.", e)
            # assume `dqm` given as file, ready for upload
            data = dqm

        logger.debug("Problem encoded for upload: %r", data)
        return data

    def sample_bqm(self, bqm, label=None, **params):
        """Use for testing."""
        from dwave.cloud.utils import bqm_to_dqm

        # to sample BQM problems, we need to convert them to DQM
        dqm = bqm_to_dqm(bqm)

        # TODO: convert sampleset back
        return self.sample_dqm(dqm, label=label, **params)

    def sample_dqm(self, dqm, label=None, **params):
        """Sample from the specified :term:`DQM`.

        Args:
            dqm (:class:`~dimod.DiscreteQuadraticModel`/str):
                A discrete quadratic model, or a reference to one
                (Problem ID returned by `.upload_dqm` method).

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Note:
            To use this method, dimod package has to be installed.
        """
        return self.sample_problem(dqm, label=label, **params)

    def upload_dqm(self, dqm):
        """Upload the specified :term:`DQM` to SAPI, returning a Problem ID
        that can be used to submit the DQM to this solver (i.e. call the
        `.sample_dqm` method).

        Args:
            dqm (:class:`~dimod.DiscreteQuadraticModel`/bytes-like/file-like):
                A discrete quadratic model given either as an in-memory
                :class:`~dimod.DiscreteQuadraticModel` object, or as raw data
                (encoded serialized model) in either a file-like or a bytes-like
                object.

        Returns:
            :class:`concurrent.futures.Future`[str]:
                Problem ID in a Future. Problem ID can be used to submit
                problems by reference.

        Note:
            To use this method, dimod package has to be installed.
        """
        return self.upload_problem(dqm)


class CQMSolver(BaseUnstructuredSolver):
    """Class for D-Wave unstructured constrained quadratic model solvers.

    This class provides a :term:`CQM` sampling method and encapsulates the
    solver description returned from the D-Wave cloud API.

    Args:
        client (:class:`~dwave.cloud.client.Client`):
            Client that manages access to this solver.

        data (`dict`):
            Data from the server describing this solver.

    Note:
        Events are not yet dispatched from unstructured solvers.
    """

    _handled_problem_types = {"cqm"}
    _handled_encoding_formats = {"bq"}

    def _encode_problem_for_upload(self, cqm):
        try:
            data = cqm.to_file()
        except Exception as e:
            logger.debug("CQM conversion to file failed with %r, "
                         "assuming data already encoded.", e)
            # assume `cqm` given as file, ready for upload
            data = cqm

        logger.debug("Problem encoded for upload: %r", data)
        return data

    def sample_bqm(self, bqm, label=None, **params):
        """Use for testing."""
        try:
            import dimod
        except ImportError: # pragma: no cover
            raise RuntimeError("Can't sample from 'cqm' without dimod. "
                               "Re-install the library with 'cqm' support.")

        # to sample BQM problems, we need to set them as objective of a CQM
        cqm = dimod.CQM.from_bqm(bqm)

        return self.sample_cqm(cqm, label=label, **params)

    def sample_cqm(self, cqm, label=None, **params):
        """Sample from the specified :term:`CQM`.

        Args:
            cqm (:class:`~dimod.ConstrainedQuadraticModel`/str):
                A constrained quadratic model, or a reference to one
                (Problem ID returned by `.upload_cqm` method).

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Note:
            To use this method, dimod package has to be installed.
        """
        return self.sample_problem(cqm, label=label, **params)

    def upload_cqm(self, cqm):
        """Upload the specified :term:`CQM` to SAPI, returning a Problem ID
        that can be used to submit the CQM to this solver (i.e. call the
        `.sample_cqm` method).

        Args:
            cqm (:class:`~dimod.ConstrainedQuadraticModel`/bytes-like/file-like):
                A constrained quadratic model given either as an in-memory
                :class:`~dimod.ConstrainedQuadraticModel` object, or as raw data
                (encoded serialized model) in either a file-like or a bytes-like
                object.

        Returns:
            :class:`concurrent.futures.Future`[str]:
                Problem ID in a Future. Problem ID can be used to submit
                problems by reference.

        Note:
            To use this method, dimod package has to be installed.
        """
        return self.upload_problem(cqm)


class StructuredSolver(BaseSolver):
    """Class for D-Wave structured solvers.

    This class provides :term:`Ising`, :term:`QUBO` and :term:`BQM` sampling
    methods and encapsulates the solver description returned from the D-Wave
    cloud API.

    Args:
        client (:class:`~dwave.cloud.client.Client`):
            Client that manages access to this solver.

        data (`dict`):
            Data from the server describing this solver.

    """

    _handled_problem_types = {"ising", "qubo"}
    _handled_encoding_formats = {"qp"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # The exact sequence of nodes/edges is used in encoding problems and must be preserved
        try:
            self._encoding_qubits = self.properties['qubits']
        except KeyError:
            raise SolverPropertyMissingError("Missing solver property: 'properties.qubits'")

        try:
            self._encoding_couplers = [tuple(edge) for edge in self.properties['couplers']]
        except KeyError:
            raise SolverPropertyMissingError("Missing solver property: 'properties.couplers'")

        # The nodes in this solver's graph: set(int)
        self.nodes = self.variables = set(self._encoding_qubits)

        # The edges in this solver's graph, every edge will be present as (a, b) and (b, a): set(tuple(int, int))
        self.edges = self.couplers = set(tuple(edge) for edge in self._encoding_couplers) | \
            set((edge[1], edge[0]) for edge in self._encoding_couplers)

        # The edges in this solver's graph, each edge will only be represented once: set(tuple(int, int))
        self.undirected_edges = {edge for edge in self.edges if edge[0] < edge[1]}

        # Create a set of default parameters for the queries
        self._params = {}

        # Add derived properties specific for this solver
        self.derived_properties.update({'lower_noise', 'num_active_qubits'})

    # Derived properties

    @property
    def num_active_qubits(self):
        "The number of active (encoding) qubits."
        return len(self.nodes)

    @property
    def is_vfyc(self):
        "Is this a virtual full-yield chip?"
        return self.properties.get('vfyc') == True

    @property
    def has_flux_biases(self):
        "Solver supports/accepts ``flux_biases``."
        return 'flux_biases' in self.parameters

    @property
    def has_anneal_schedule(self):
        "Solver supports/accepts ``anneal_schedule``."
        return 'anneal_schedule' in self.parameters

    @property
    def num_qubits(self):
        "Nominal number of qubits on chip (includes active AND inactive)."
        return self.properties.get('num_qubits')

    @property
    def lower_noise(self):
        return "lower_noise" in self.properties.get("tags", [])

    def max_num_reads(self, **params):
        """Returns the maximum number of reads for the given solver parameters.

        Args:
            **params:
                Parameters for the sampling method. Relevant to num_reads:

                - annealing_time
                - readout_thermalization
                - num_reads
                - programming_thermalization

        Returns:
            int: The maximum number of reads.

        """
        # dev note: in the future it would be good to have a way of doing this
        # server-side, as we are duplicating logic here.

        properties = self.properties

        if self.software or not params:
            # software solvers don't use any of the above parameters
            return properties['num_reads_range'][1]

        # qpu

        _, duration = properties['problem_run_duration_range']

        annealing_time = params.get('annealing_time',
                                    properties['default_annealing_time'])

        readout_thermalization = params.get('readout_thermalization',
                                            properties['default_readout_thermalization'])

        programming_thermalization = params.get('programming_thermalization',
                                                properties['default_programming_thermalization'])

        return min(properties['num_reads_range'][1],
                   int((duration - programming_thermalization)
                       / (annealing_time + readout_thermalization)))

    # Sampling methods

    def sample_ising(self, linear, quadratic, offset=0, label=None, **params):
        """Sample from the specified :term:`Ising` model.

        Args:
            linear (dict/list):
                Linear biases of the Ising problem. If a dict, should be of the
                form `{v: bias, ...}` where v is a spin-valued variable and `bias`
                is its associated bias. If a list, it is treated as a list of
                biases where the indices are the variable labels.

            quadratic (dict[(int, int), float]):
                Quadratic terms of the model (J), stored in a dict. With keys
                that are 2-tuples of variables and values are quadratic biases
                associated with the pair of variables (the interaction).

            offset (float, optional, default=0):
                Constant offset applied to the model.

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Examples:
            This example creates a client using the local system's default D-Wave Cloud Client
            configuration file, which is configured to access a D-Wave 2000Q QPU, submits a
            simple :term:`Ising` problem (opposite linear biases on two coupled qubits), and samples
            5 times.

            >>> from dwave.cloud import Client
            >>> with Client.from_config() as client:
            ...     solver = client.get_solver()
            ...     u, v = next(iter(solver.edges))
            ...     computation = solver.sample_ising({u: -1, v: 1}, {}, num_reads=5)   # doctest: +SKIP
            ...     for i in range(5):
            ...         print(computation.samples[i][u], computation.samples[i][v])
            ...
            ...
            (1, -1)
            (1, -1)
            (1, -1)
            (1, -1)
            (1, -1)

        """
        # Our linear and quadratic objective terms are already separated in an
        # ising model so we can just directly call `_sample`.
        return self._sample('ising', linear, quadratic, offset, params, label=label)

    def sample_qubo(self, qubo, offset=0, label=None, **params):
        """Sample from the specified :term:`QUBO`.

        Args:
            qubo (dict[(int, int), float]):
                Coefficients of a quadratic unconstrained binary optimization
                (QUBO) problem. Should be a dict of the form `{(u, v): bias, ...}`
                where `u`, `v`, are binary-valued variables and `bias` is their
                associated coefficient.

            offset (optional, default=0):
                Constant offset applied to the model.

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Examples:
            This example creates a client using the local system's default D-Wave Cloud Client
            configuration file, which is configured to access a D-Wave 2000Q QPU, submits
            a :term:`QUBO` problem (a Boolean NOT gate represented by a penalty model), and
            samples 5 times.

            >>> from dwave.cloud import Client
            >>> with Client.from_config() as client:  # doctest: +SKIP
            ...     solver = client.get_solver()
            ...     u, v = next(iter(solver.edges))
            ...     Q = {(u, u): -1, (u, v): 0, (v, u): 2, (v, v): -1}
            ...     computation = solver.sample_qubo(Q, num_reads=5)
            ...     for i in range(5):
            ...         print(computation.samples[i][u], computation.samples[i][v])
            ...
            ...
            (0, 1)
            (1, 0)
            (1, 0)
            (0, 1)
            (1, 0)

        """
        linear, quadratic = reformat_qubo_as_ising(qubo)
        return self._sample('qubo', linear, quadratic, offset, params, label=label)

    def sample_bqm(self, bqm, label=None, **params):
        """Sample from the specified :term:`BQM`.

        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`):
                A binary quadratic model.

            label (str, optional):
                Problem label you can optionally tag submissions with for ease
                of identification.

            **params:
                Parameters for the sampling method, solver-specific.

        Returns:
            :class:`~dwave.cloud.computation.Future`

        Note:
            To use this method, dimod package has to be installed.
        """
        try:
            import dimod
        except ImportError: # pragma: no cover
            raise RuntimeError("Can't sample from 'bqm' without dimod. "
                               "Re-install the library with 'bqm' support.")

        if bqm.vartype is dimod.SPIN:
            problem_type = 'ising'
        elif bqm.vartype is dimod.BINARY:
            problem_type = 'qubo'
        else:
            raise TypeError("unknown/unsupported vartype")

        return self._sample(problem_type, bqm.linear, bqm.quadratic, bqm.offset,
                            params, label=label, undirected_biases=True)

    @dispatches_events('sample')
    def _sample(self, type_, linear, quadratic, offset, params,
                label=None, undirected_biases=False):
        """Internal method for `sample_ising`, `sample_qubo` and `sample_bqm`.

        Args:
            linear (list/dict):
                Linear terms of the model.

            quadratic (dict[(int, int), float]):
                Quadratic terms of the model.

            offset (number):
                Constant offset applied to the model.

            params (dict):
                Parameters for the sampling method, solver-specific.

            label (str, optional):
                Problem label.

            undirected_biases (boolean, default=False):
                Are (quadratic) biases specified on undirected edges? For
                triangular or symmetric matrix of quadratic biases set it to
                ``True``.

        Returns:
            :class:`~dwave.cloud.computation.Future`
        """

        # Check the problem
        if not self.check_problem(linear, quadratic):
            raise ProblemStructureError(
                f"Problem graph incompatible with {self.id} solver")

        # Mix the new parameters with the default parameters
        combined_params = dict(self._params)
        combined_params.update(params)

        # Check the parameters before submitting
        for key in combined_params:
            if key not in self.parameters and not key.startswith('x_'):
                raise KeyError("{} is not a parameter of this solver.".format(key))

        # transform some of the parameters in-place
        self._format_params(type_, combined_params)

        body_dict = {
            'solver': self.id,
            'data': encode_problem_as_qp(self, linear, quadratic, offset,
                                         undirected_biases=undirected_biases),
            'type': type_,
            'params': combined_params
        }
        if label is not None:
            body_dict['label'] = label
        body_data = json.dumps(body_dict, cls=NumpyEncoder)
        logger.trace("Encoded sample request: %s", body_data)

        body = Present(result=body_data)
        computation = Future(solver=self, id_=None, return_matrix=self.return_matrix)

        # XXX: offset is carried on Future until implemented in SAPI
        computation._offset = offset

        logger.debug("Submitting new problem to: %s", self.id)
        self.client._submit(body, computation)

        return computation

    # kept for internal backwards compatibility and in case it's being
    # used externally anywhere.
    def _format_params(self, type_, params):
        """Reformat some of the parameters for sapi."""
        self.reformat_parameters(type_, params, self.properties, inplace=True)

    @staticmethod
    def reformat_parameters(vartype: '_Vartype',
                            parameters: typing.MutableMapping[str, typing.Any],
                            properties: typing.Mapping[str, typing.Any],
                            inplace: bool = False,
                            ) -> typing.MutableMapping[str, typing.Any]:
        """Reformat some solver parameters for SAPI.

        Currently the only reformatted parameter is ``initial_state``. This
        method allows ``initial_state`` to be submitted as a dictionary
        mapping the qubits to their initial value.

        Args:
            vartype: One of ``'ising'`` or ``'qubo'``. If :mod:`dimod` is
                installed, this can also be any
                :class:`~dimod.typing.VartypeLike`.
            parameters: The parameters to submit to this solver.
            properties: The solver's properties. Either
                :attr:`StructuredSolver.properties` or
                :attr:`dwave.systems.DWaveSampler.properties` can be
                provided.
            inplace: Whether to modify the ``parameters`` in-place or return
                a copy.

        Returns:
            The reformatted solver parameters.
            If ``inplace`` is true the modified ``parameters`` is returned.
            If ``inplace`` is false then a deep copy of ``parameters``
            with the relevant fields updated is returned.

        """
        # whether to copy or not
        parameters = parameters if inplace else copy.deepcopy(parameters)

        # handle the vartype
        if vartype not in ('ising', 'qubo'):
            try:
                vartype = 'ising' if dimod.as_vartype(vartype) is dimod.SPIN else 'qubo'
            except (TypeError, AttributeError):
                msg = "expected vartype to be one of: 'ising', 'qubo'"
                if dimod:
                    msg += ", 'BINARY', 'SPIN', dimod.BINARY, dimod.SPIN"
                msg += f"; {vartype!r} provided"
                raise ValueError(msg) from None

        # update the parameters
        if 'initial_state' in parameters:
            initial_state = parameters['initial_state']
            if isinstance(initial_state, typing.Mapping):

                initial_state_list = [3]*properties['num_qubits']

                low = -1 if vartype == 'ising' else 0

                for v, val in initial_state.items():
                    if val == 3:
                        continue
                    if val <= 0:
                        initial_state_list[v] = low
                    else:
                        initial_state_list[v] = 1

                parameters['initial_state'] = initial_state_list
            # else: support old format

        return parameters

    def check_problem(self, linear, quadratic):
        """Test if an Ising model matches the graph provided by the solver.

        Args:
            linear (list/dict):
                Linear terms of the model (h).

            quadratic (dict[(int, int), float]):
                Quadratic terms of the model (J).

        Returns:
            bool

        Examples:
            This example creates a client using the local system's default D-Wave Cloud Client
            configuration file, which is configured to access a D-Wave 2000Q QPU, and
            tests a simple :term:`Ising` model for two target embeddings (that is, representations
            of the model's graph by coupled qubits on the QPU's sparsely connected graph),
            where only the second is valid.

            >>> from dwave.cloud import Client
            >>> print((0, 1) in solver.edges)   # doctest: +SKIP
            False
            >>> print((0, 4) in solver.edges)   # doctest: +SKIP
            True
            >>> with Client.from_config() as client:  # doctest: +SKIP
            ...     solver = client.get_solver()
            ...     print(solver.check_problem({0: -1, 1: 1},{(0, 1):0.5}))
            ...     print(solver.check_problem({0: -1, 4: 1},{(0, 4):0.5}))
            ...
            False
            True
        """
        # handle legacy format
        if not isinstance(linear, Mapping):
             linear = {idx: val for idx, val in enumerate(linear) if val != 0}

        return self.nodes.issuperset(linear) and self.edges.issuperset(quadratic)

    def estimate_qpu_access_time(self,
                                 num_qubits: int,
                                 num_reads: int = 1,
                                 annealing_time: typing.Optional[float] = None,
                                 anneal_schedule: typing.Optional[typing.List[typing.Tuple[float, float]]] = None,
                                 initial_state:  typing.Optional[typing.List[typing.Tuple[float, float]]] = None,
                                 reverse_anneal: bool = False,
                                 reinitialize_state: bool = False,
                                 programming_thermalization: typing.Optional[float] = None,
                                 readout_thermalization: typing.Optional[float] = None,
                                 reduce_intersample_correlation: bool = False,
                                 **kwargs) -> float:
        """Estimates QPU access time for a submission to the selected solver.

        Estimates a problem’s quantum processing unit (QPU) access time from the
        parameter values you specify, timing data provided in the ``problem_timing_data``
        `solver property <https://docs.dwavesys.com/docs/latest/c_solver_properties.html>`_,
        and the number of qubits used to embed the problem on the selected QPU, as
        described in the
        `system documentation <https://docs.dwavesys.com/docs/latest/c_qpu_timing.html>`_.

        Requires that you provide the number of qubits to be used for your
        problem submission. :term:`Embedding` is typically heuristic and the number
        of required qubits can vary between executions.

        Args:
            num_qubits:
                Number of qubits required to represent your binary quadratic model
                on the selected solver.
            num_reads:
                Number of reads. Provide this value if you explictly set ``num_reads``
                in your submission.
            annealing_time:
                Annealing duration. Provide this value of if you set
                ``annealing_time`` in your submission.
            anneal_schedule:
                Anneal schedule. Provide the ``anneal_schedule`` if you set it in
                your submission.
            initial_state:
                Initial state. Provide the ``initial_state`` if your submission
                uses reverse annealing.
            reinitialize_state:
                Set to ``True`` if your submission sets ``reinitialize_state``.
            programming_thermalization:
                programming thermalization time. Provide this value if you explictly
                set a value for ``programming_thermalization`` in your submission.
            readout_thermalization:
                Set to ``True`` if your submission sets ``readout_thermalization``.
            reduce_intersample_correlation:
                Set to ``True`` if your submission sets ``reduce_intersample_correlation``.

        Returns:
            Estimated QPU access time, in microseconds.

        Raises:
            KeyError: If a solver property, or a field in the ``problem_timing_data``
                solver property, required by the timing model is missing for the
                selected solver.

            ValueError: If conflicting parameters are set or the selected solver
                uses an unsupported timing model.

        Examples:

            This example estimates the QPU access time for a ferromagnetic problem
            using all the selected QPU's qubits before deciding whether to submit
            the problem.

            .. testsetup::

                estimated_runtime = 42657   # to test solver.sample_bqm()

            >>> from dimod import BinaryQuadraticModel
            >>> from dwave.cloud import Client
            >>> reads = 100
            >>> max_time = 100000
            >>> with Client.from_config() as client:
            ...    solver = client.get_solver(qpu=True)
            ...    bqm = BinaryQuadraticModel({}, {edge: -1 for edge in solver.edges}, "BINARY")
            ...    num_qubits = len(solver.nodes)
            ...    estimated_runtime = solver.estimate_qpu_access_time(num_qubits, num_reads=reads) # doctest: +SKIP
            ...    print("Estimate of {:.0f}us on {}".format(estimated_runtime, solver.name)) # doctest: +SKIP
            ...    if estimated_runtime < max_time:
            ...       computation = solver.sample_bqm(bqm, num_reads=reads)
            Estimate of 42657us on Advantage_system4.1
            >>> print("QPU access time: {:.0f}us".format(computation.timing["qpu_access_time"]))  # doctest: +SKIP
            QPU access time: 42640us

        """
        if anneal_schedule and annealing_time:
            raise ValueError("set only one of ``anneal_schedule`` or ``annealing_time``")

        if anneal_schedule and anneal_schedule[0][1] == 1 and not initial_state:
            raise ValueError("reverse annealing requested (``anneal_schedule`` "
                             "starts with s==1) but ``initial_state`` not provided")

        try:
            problem_timing_data = self.properties['problem_timing_data']
        except:
            raise KeyError("selected solver is missing required property ``problem_timing_data``")

        try:
            version_timing_model = problem_timing_data['version']
        except:
            raise KeyError("selected solver is missing ``problem_timing_data`` field ``version``")

        try:
            typical_programming_time = problem_timing_data['typical_programming_time']
            ra_with_reinit_prog_time_delta = problem_timing_data['reverse_annealing_with_reinit_prog_time_delta']
            ra_without_reinit_prog_time_delta = problem_timing_data['reverse_annealing_without_reinit_prog_time_delta']
            default_programming_thermalization = problem_timing_data['default_programming_thermalization']
            default_annealing_time = problem_timing_data['default_annealing_time']
            readout_time_model = problem_timing_data['readout_time_model']
            readout_time_model_parameters = problem_timing_data['readout_time_model_parameters']
            qpu_delay_time_per_sample = problem_timing_data['qpu_delay_time_per_sample']
            ra_with_reinit_delay_time_delta = problem_timing_data['reverse_annealing_with_reinit_delay_time_delta']
            ra_without_reinit_delay_time_delta = problem_timing_data['reverse_annealing_without_reinit_delay_time_delta']
            decorrelation_max_nominal_anneal_time = problem_timing_data['decorrelation_max_nominal_anneal_time']
            decorrelation_time_range = problem_timing_data['decorrelation_time_range']
            default_readout_thermalization = problem_timing_data['default_readout_thermalization']

        except KeyError as err:
            err_msg = f"selected solver is missing ``problem_timing_data`` field " + \
                      f"{err} required by timing model version {version_timing_model}"
            raise ValueError(err_msg)

        # Support for sapi timing model versions: 1.0.x
        if not version_timing_model.startswith("1.0."):
            raise ValueError(f"``estimate_qpu_access_time`` does not currently "
                             f"support timing model {version_timing_model} "
                             "used by the selected solver")

        ra_programming_time = 0
        ra_delay_time = 0
        if anneal_schedule and anneal_schedule[0][1] == 1:
            if reinitialize_state:
                ra_programming_time = ra_with_reinit_prog_time_delta
                ra_delay_time = ra_with_reinit_delay_time_delta
            else:
                ra_programming_time = ra_without_reinit_prog_time_delta
                ra_delay_time = ra_without_reinit_delay_time_delta

        if programming_thermalization:
            programming_thermalization_time = programming_thermalization
        else:
            programming_thermalization_time = default_programming_thermalization

        programming_time = typical_programming_time + ra_programming_time + \
                           programming_thermalization_time

        anneal_time = default_annealing_time
        if annealing_time:
            anneal_time = annealing_time
        elif anneal_schedule:
            anneal_time = anneal_schedule[-1][0]

        n = len(readout_time_model_parameters)
        if n % 2:
             raise ValueError(f"for the selected solver, ``problem_timing_data`` "
                              f"property field ``readout_time_model_parameters`` "
                              f"is not of an even length as required by timing "
                              f"model version {version_timing_model}")

        q = readout_time_model_parameters[:n//2]
        t = readout_time_model_parameters[n//2:]
        if readout_time_model == 'pwl_log_log':
            readout_time = pow(10, np.interp(np.emath.log10(num_qubits), q, t))
        elif readout_time_model == 'pwl_linear':
            readout_time = np.interp(num_qubits, q, t)
        else:
            raise ValueError("``estimate_qpu_access_time`` does not support "
                             f"``readout_time_model`` value {readout_time_model} "
                             f"in version {version_timing_model}")

        if readout_thermalization:
            readout_thermalization_time = readout_thermalization
        else:
            readout_thermalization_time = default_readout_thermalization

        decorrelation_time = 0
        if reduce_intersample_correlation:
            r_min = decorrelation_time_range[0]
            r_max = decorrelation_time_range[1]
            decorrelation_time = anneal_time/decorrelation_max_nominal_anneal_time * (r_max - r_min) + r_min

        sampling_time_per_read = anneal_time + readout_time + qpu_delay_time_per_sample + \
            ra_delay_time + max(readout_thermalization_time, decorrelation_time)

        sampling_time = num_reads * sampling_time_per_read

        return sampling_time + programming_time


# for backwards compatibility:
Solver = StructuredSolver
UnstructuredSolver = BQMSolver

# list of all available solvers, ordered according to loading attempt priority
# (more specific first)
available_solvers = [StructuredSolver, BQMSolver, CQMSolver, DQMSolver]



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union, Optional
from datetime import datetime

import numpy
from pydantic import BaseModel

from dwave.cloud.api import constants


class BaseModelWithEncoders(BaseModel):
    class Config:
        json_encoders = {
            numpy.integer: int,
            numpy.floating: float,
            numpy.bool_: bool,
            numpy.ndarray: lambda obj: obj.tolist(),
        }


class SolverConfiguration(BaseModel):
    id: str
    status: str
    description: str
    properties: dict
    avg_load: float


class ProblemInitialStatus(BaseModel):
    id: str
    type: constants.ProblemType
    solver: str
    label: Optional[str]
    status: constants.ProblemStatus
    submitted_on: datetime


class ProblemStatus(ProblemInitialStatus):
    solved_on: Optional[datetime]


class StructuredProblemAnswer(BaseModel):
    format: constants.AnswerEncodingFormat = constants.AnswerEncodingFormat.QP
    active_variables: str
    energies: str
    solutions: str
    timing: dict
    num_occurrences: str
    num_variables: int


class UnstructuredProblemAnswer(BaseModel):
    format: constants.AnswerEncodingFormat = constants.AnswerEncodingFormat.BQ
    data: dict


class ProblemAnswer(BaseModel):
    __root__: Union[StructuredProblemAnswer, UnstructuredProblemAnswer]

    def __getattr__(self, item):
        return getattr(self.__root__, item)

    def dict(self, **kwargs):
        return super().dict(**kwargs)['__root__']


class ProblemStatusWithAnswer(ProblemStatus):
    answer: ProblemAnswer


class ProblemStatusMaybeWithAnswer(ProblemStatus):
    answer: Optional[ProblemAnswer]


class StructuredProblemData(BaseModel):
    format: constants.ProblemEncodingFormat = constants.ProblemEncodingFormat.QP
    lin: str
    quad: str
    offset: float = 0.0


class UnstructuredProblemData(BaseModel):
    format: constants.ProblemEncodingFormat = constants.ProblemEncodingFormat.REF
    data: str


class ProblemData(BaseModel):
    __root__: Union[StructuredProblemData, UnstructuredProblemData]

    def __getattr__(self, item):
        return getattr(self.__root__, item)

    def dict(self, **kwargs):
        return super().dict(**kwargs)['__root__']


class ProblemMetadata(BaseModel):
    solver: str
    type: constants.ProblemType
    label: Optional[str]
    status: constants.ProblemStatus
    submitted_by: str
    submitted_on: datetime
    solved_on: Optional[datetime]
    messages: Optional[List[dict]]


class ProblemInfo(BaseModel):
    id: str
    data: ProblemData
    params: dict
    metadata: ProblemMetadata
    answer: ProblemAnswer


class ProblemJob(BaseModelWithEncoders):
    data: ProblemData
    params: dict
    solver: str
    type: constants.ProblemType
    label: Optional[str]


class BatchItemError(BaseModel):
    error_code: int
    error_msg: str

class ProblemSubmitError(BatchItemError):
    pass

class ProblemCancelError(BatchItemError):
    pass


# region info on metadata api
class Region(BaseModel):
    code: str
    name: str
    endpoint: str



# Copyright 2017 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test problem submission against hard-coded replies with unittest.mock."""

import json
import unittest
import threading
import collections

from unittest import mock
from requests.structures import CaseInsensitiveDict
from requests.exceptions import HTTPError
from concurrent.futures import TimeoutError
from parameterized import parameterized

import numpy

try:
    import dimod
except ImportError:
    dimod = None

from dwave.cloud.client import Client
from dwave.cloud.solver import Solver
from dwave.cloud.computation import Future
from dwave.cloud.utils import evaluate_ising, utcrel
from dwave.cloud.exceptions import (
    SolverFailureError, CanceledFutureError, SolverError,
    InvalidAPIResponseError)

from tests.api.mocks import StructuredSapiMockResponses


def choose_reply(path, replies, statuses=None, date=None):
    """Choose the right response based on the path and make a mock response."""

    if statuses is None:
        statuses = collections.defaultdict(lambda: iter([200]))

    if date is None:
        date = utcrel(0)

    if path in replies:
        response = mock.Mock(['text', 'json', 'raise_for_status', 'headers'])
        response.status_code = next(statuses[path])
        text = replies[path]
        if not isinstance(text, str):
            text = json.dumps(text)
        response.text = text
        response.json.side_effect = lambda: replies[path]
        response.headers = CaseInsensitiveDict({'Date': date.isoformat()})

        def raise_for_status():
            if not 200 <= response.status_code < 400:
                raise HTTPError(response.status_code)
        response.raise_for_status = raise_for_status

        def ok():
            try:
                response.raise_for_status()
            except HTTPError:
                return False
            return True
        ok_property = mock.PropertyMock(side_effect=ok)
        type(response).ok = ok_property

        return response
    else:
        raise NotImplementedError(path)


class _QueryTest(unittest.TestCase):
    def _check(self, results, linear, quad, offset=0, num_reads=1):
        # Did we get the right number of samples?
        self.assertEqual(num_reads, sum(results.num_occurrences))

        # verify num_occurrences sum corresponds to num_reads
        self.assertEqual(100, sum(results.num_occurrences))

        # Make sure energies are correct in raw results
        for energy, state in zip(results.energies, results.samples):
            self.assertEqual(energy, evaluate_ising(linear, quad, state, offset=offset))

        # skip sampleset test if dimod is not installed
        if not dimod:
            return

        # Make sure the sampleset matches raw results
        for record, energy, num_occurrences, state in \
                zip(results.sampleset.record,
                    results['energies'],
                    results['num_occurrences'],
                    results['solutions']):
            recalc_energy = evaluate_ising(linear, quad, state, offset=offset)
            self.assertEqual(energy, recalc_energy)
            self.assertEqual(energy, float(record.energy))
            self.assertEqual(num_occurrences, int(record.num_occurrences))
            self.assertEqual(state, list(record.sample))


@mock.patch('time.sleep', lambda *x: None)
class MockSubmission(_QueryTest):
    """Test connecting and some related failure modes."""

    @classmethod
    def setUpClass(cls):
        cls.sapi = StructuredSapiMockResponses()

    def test_submit_null_reply(self):
        """Get an error when the server's response is incomplete."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': ''})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                results = solver.sample_ising(linear, quadratic)

                with self.assertRaises(InvalidAPIResponseError):
                    results.samples

    def test_submit_ising_ok_reply(self):
        """Handle a normal query and response."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(id='123')})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)
                results = solver.sample_ising(linear, quadratic, **params)

                self._check(results, linear, quadratic, **params)

    @unittest.skipUnless(dimod, "dimod required for 'Solver.sample_bqm'")
    def test_submit_bqm_ising_ok_reply(self):
        """Handle a normal query and response."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(id='123')})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                h, J = self.sapi.problem
                bqm = dimod.BinaryQuadraticModel.from_ising(h, J)

                params = dict(num_reads=100)
                results = solver.sample_bqm(bqm, **params)

                self._check(results, h, J, **params)

    def test_submit_qubo_ok_reply(self):
        """Handle a normal query and response."""

        qubo_msg_diff = dict(type="qubo")
        qubo_answer_diff = {
            'energies': 'AAAAAAAAAAA=',
            'solutions': 'AA==',
            'active_variables': 'AAAAAAQAAAA='
        }

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(
                    id='123', answer_patch=qubo_answer_diff, **qubo_msg_diff)})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                qubo = {(0, 0): 4.0, (0, 4): -4, (4, 4): 4.0}
                offset = -2.0
                params = dict(num_reads=100)

                results = solver.sample_qubo(qubo, offset, **params)

                # make sure energies are correct in raw results
                for energy, sample in zip(results.energies, results.samples):
                    self.assertEqual(energy, evaluate_ising({}, qubo, sample, offset=offset))

    @unittest.skipUnless(dimod, "dimod required for 'Solver.sample_bqm'")
    def test_submit_bqm_qubo_ok_reply(self):
        """Handle a normal query and response."""

        qubo_msg_diff = dict(type="qubo")
        qubo_answer_diff = {
            'energies': 'AAAAAAAAAAA=',
            'solutions': 'AA==',
            'active_variables': 'AAAAAAQAAAA='
        }

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(
                    id='123', answer_patch=qubo_answer_diff, **qubo_msg_diff)})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                qubo = {(0, 0): 4.0, (0, 4): -4, (4, 4): 4.0}
                offset = -2.0
                params = dict(num_reads=100)

                results = solver.sample_qubo(qubo, offset, **params)

                # make sure energies are correct in raw results
                for energy, sample in zip(results.energies, results.samples):
                    self.assertEqual(energy, evaluate_ising({}, qubo, sample, offset=offset))

    def test_submit_error_reply(self):
        """Handle an error on problem submission."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.error_reply(error_message='An error message')]})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                results = solver.sample_ising(linear, quadratic)

                with self.assertRaises(SolverFailureError):
                    results.samples

    def test_submit_immediate_error_reply(self):
        """Handle an (obvious) error on problem submission."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.immediate_error_reply(
                    code=400, msg="Missing parameter 'num_reads' in problem JSON")]})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                results = solver.sample_ising(linear, quadratic)

                # waiting should be canceled on exception (i.e. NOT timeout)
                self.assertTrue(results.wait(timeout=1))
                self.assertIsNone(results.wait_id(timeout=1))

                # resolving result should raise exception
                with self.assertRaises(SolverFailureError):
                    results.samples

    def test_submit_cancel_reply(self):
        """Handle a response for a canceled job."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.cancel_reply()]})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                results = solver.sample_ising(linear, quadratic)

                with self.assertRaises(CanceledFutureError):
                    results.samples

    def test_answer_load_error(self):
        """Answer load error is propagated as exception."""

        error_code = 404
        error_message = 'Problem not found'

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda path, _: choose_reply(path, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda path: choose_reply(
                path, replies={
                    'problems/123/': error_message
                }, statuses={
                    'problems/123/': iter([error_code])
                })
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                future = solver.sample_ising(linear, quadratic)

                with self.assertRaises(SolverError) as exc:
                    future.result()

                self.assertEqual(str(exc.exception), error_message)

    def test_submit_continue_then_ok_reply(self):
        """Handle polling for a complete problem."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.continue_reply(id='123')]
            })
            session.get = lambda a: choose_reply(a, {
                'problems/?id=123': [self.sapi.complete_no_answer_reply(id='123')],
                'problems/123/': self.sapi.complete_reply(id='123')
            })
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)
                results = solver.sample_ising(linear, quadratic, **params)

                self._check(results, linear, quadratic, **params)

    def test_submit_continue_then_error_reply(self):
        """Handle polling for an error message."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.continue_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/?id=123': [self.sapi.error_reply(id='123')]})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)
                results = solver.sample_ising(linear, quadratic, **params)

                with self.assertRaises(SolverFailureError):
                    self._check(results, linear, quadratic, **params)

    def test_submit_continue_then_ok_and_error_reply(self):
        """Handle polling for the status of multiple problems."""

        # we need a "global session", because mocked responses are stateful
        def global_mock_session():
            session = mock.Mock()

            # on first status poll, return pending for both problems
            # on second status poll, return error for first problem and complete for second
            def continue_then_complete(path, state={'count': 0}):
                state['count'] += 1
                if state['count'] < 2:
                    return choose_reply(path, {
                        'problems/?id=1': [self.sapi.continue_reply(id='1')],
                        'problems/?id=2': [self.sapi.continue_reply(id='2')],
                        'problems/1/': self.sapi.continue_reply(id='1'),
                        'problems/2/': self.sapi.continue_reply(id='2'),
                        'problems/?id=1,2': [self.sapi.continue_reply(id='1'),
                                             self.sapi.continue_reply(id='2')],
                        'problems/?id=2,1': [self.sapi.continue_reply(id='2'),
                                             self.sapi.continue_reply(id='1')]
                    })
                else:
                    return choose_reply(path, {
                        'problems/?id=1': [self.sapi.error_reply(id='1')],
                        'problems/?id=2': [self.sapi.complete_no_answer_reply(id='2')],
                        'problems/1/': self.sapi.error_reply(id='1'),
                        'problems/2/': self.sapi.complete_reply(id='2'),
                        'problems/?id=1,2': [self.sapi.error_reply(id='1'),
                                             self.sapi.complete_no_answer_reply(id='2')],
                        'problems/?id=2,1': [self.sapi.complete_no_answer_reply(id='2'),
                                             self.sapi.error_reply(id='1')]
                    })

            def accept_problems_with_continue_reply(path, body, ids=iter('12')):
                problems = json.loads(body)
                return choose_reply(path, {
                    'problems/': [self.sapi.continue_reply(id=next(ids)) for _ in problems]
                })

            session.get = continue_then_complete
            session.post = accept_problems_with_continue_reply

            return session

        session = global_mock_session()

        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)

                results1 = solver.sample_ising(linear, quadratic, **params)
                results2 = solver.sample_ising(linear, quadratic, **params)

                with self.assertRaises(SolverFailureError):
                    self._check(results1, linear, quadratic, **params)
                self._check(results2, linear, quadratic, **params)

    def test_exponential_backoff_polling(self):
        "After each poll, back-off should double"

        # we need a "global session", because mocked responses are stateful
        def global_mock_session():
            session = mock.Mock()

            # on submit, return status pending
            session.post = lambda path, _: choose_reply(path, {
                'problems/': [self.sapi.continue_reply(id='123')]
            })

            # on first and second status poll, return pending
            # on third status poll, return completed
            def continue_then_complete(path, state={'count': 0}):
                state['count'] += 1
                if state['count'] < 3:
                    return choose_reply(path, {
                        'problems/?id=123': [self.sapi.continue_reply(id='123')],
                        'problems/123/': self.sapi.continue_reply(id='123')
                    })
                else:
                    return choose_reply(path, {
                        'problems/?id=123': [self.sapi.complete_no_answer_reply(id='123')],
                        'problems/123/': self.sapi.complete_reply(id='123')
                    })

            session.get = continue_then_complete

            return session

        session = global_mock_session()

        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                future = solver.sample_qubo({})
                future.result()

                # after third poll, back-off interval should be 4 x initial back-off
                self.assertAlmostEqual(
                    future._poll_backoff,
                    client.poll_backoff_min * client.poll_backoff_base**2)

    def test_immediate_polling(self):
        "First poll happens with minimal delay"

        # each thread can have its instance of a session because
        # responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda path, _: choose_reply(path, {
                'problems/': [self.sapi.continue_reply(id='1')]
            })
            session.get = lambda path: choose_reply(path, {
                'problems/?id=1': [self.sapi.complete_no_answer_reply(id='1')],
                'problems/1/': self.sapi.complete_reply(id='1')
            })
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                def assert_no_delay(s):
                    s and self.assertTrue(
                        abs(s - client.poll_backoff_min) < client.DEFAULTS['poll_backoff_min'])

                with mock.patch('time.sleep', assert_no_delay):
                    future = solver.sample_qubo({})
                    future.result()

    def test_immediate_polling_with_local_clock_unsynced(self):
        """First poll happens with minimal delay if local clock is way off from
        the remote/server clock."""

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            badnow = utcrel(100)
            session = mock.Mock()
            session.post = lambda path, _: choose_reply(path, {
                'problems/': [self.sapi.continue_reply(id='1')]
            }, date=badnow)
            session.get = lambda path: choose_reply(path, {
                'problems/?id=1': [self.sapi.complete_no_answer_reply(id='1')],
                'problems/1/': self.sapi.complete_reply(id='1')
            }, date=badnow)
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                def assert_no_delay(s):
                    s and self.assertTrue(
                        abs(s - client.poll_backoff_min) < client.DEFAULTS['poll_backoff_min'])

                with mock.patch('time.sleep', assert_no_delay):
                    future = solver.sample_qubo({})
                    future.result()

    def test_polling_recovery_after_5xx(self):
        "Polling shouldn't be aborted on 5xx responses."

        # we need a "global session", because mocked responses are stateful
        def global_mock_session():
            session = mock.Mock()

            # on submit, return status pending
            session.post = lambda path, _: choose_reply(path, {
                'problems/': [self.sapi.continue_reply(id='123')]
            })

            # on first and second status poll, fail with 503 and 504
            # on third status poll, return completed
            statuses = iter([503, 504])
            def continue_then_complete(path, state={'count': 0}):
                state['count'] += 1
                if state['count'] < 3:
                    return choose_reply(path, replies={
                        'problems/?id=123': [self.sapi.continue_reply(id='123')],
                        'problems/123/': self.sapi.continue_reply(id='123')
                    }, statuses={
                        'problems/?id=123': statuses,
                        'problems/123/': statuses
                    })
                else:
                    return choose_reply(path, {
                        'problems/?id=123': [self.sapi.complete_no_answer_reply(id='123')],
                        'problems/123/': self.sapi.complete_reply(id='123')
                    })

            session.get = continue_then_complete

            return session

        session = global_mock_session()

        with mock.patch.object(Client, 'create_session', lambda self: session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                future = solver.sample_qubo({})
                future.result()

                # after third poll, back-off interval should be 4 x initial back-off
                self.assertAlmostEqual(
                    future._poll_backoff,
                    client.poll_backoff_min * client.poll_backoff_base**2)


class DeleteEvent(Exception):
    """Throws exception when mocked client submits an HTTP DELETE request."""

    def __init__(self, url, body):
        """Return the URL of the request with the exception for test verification."""
        self.url = url
        self.body = body

    @staticmethod
    def handle(path, **kwargs):
        """Callback useable to mock a delete request."""
        raise DeleteEvent(path, json.dumps(kwargs['json']))


@mock.patch('time.sleep', lambda *x: None)
class MockCancel(unittest.TestCase):
    """Make sure cancel works at the two points in the process where it should."""

    @classmethod
    def setUpClass(cls):
        cls.sapi = StructuredSapiMockResponses()

    def test_cancel_with_id(self):
        """Make sure the cancel method submits to the right endpoint.

        When cancel is called after the submission is finished.
        """
        submission_id = 'test-id'

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            reply_body = [self.sapi.continue_reply(id=submission_id, solver='solver')]
            session.get = lambda a: choose_reply(a, {
                'problems/?id={}'.format(submission_id): reply_body})
            session.delete = DeleteEvent.handle
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)
                future = solver._retrieve_problem(submission_id)
                future.cancel()

                try:
                    self.assertTrue(future.id is not None)
                    future.samples
                    self.fail()
                except DeleteEvent as event:
                    if event.url == 'problems/':
                        self.assertEqual(event.body, '["{}"]'.format(submission_id))
                    else:
                        self.assertEqual(event.url, 'problems/{}/'.format(submission_id))

    def test_cancel_without_id(self):
        """Make sure the cancel method submits to the right endpoint.

        When cancel is called before the submission has returned the problem id.
        """
        submission_id = 'test-id'
        release_reply = threading.Event()

        # each thread can have its instance of a session because
        # we use a global lock (event) in the mocked responses
        def create_mock_session(client):
            reply_body = [self.sapi.continue_reply(id=submission_id)]

            session = mock.Mock()
            session.get = lambda a: choose_reply(a, {
                'problems/?id={}'.format(submission_id): reply_body})

            def post(a, _):
                release_reply.wait()
                return choose_reply(a, {'problems/': reply_body})

            session.post = post
            session.delete = DeleteEvent.handle

            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem

                future = solver.sample_ising(linear, quadratic)
                future.cancel()

                try:
                    release_reply.set()
                    future.samples
                    self.fail()
                except DeleteEvent as event:
                    if event.url == 'problems/':
                        self.assertEqual(event.body, '["{}"]'.format(submission_id))
                    else:
                        self.assertEqual(event.url, 'problems/{}/'.format(submission_id))


class TestComputationID(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.sapi = StructuredSapiMockResponses()

    def test_id_getter_setter(self):
        """Future.get_id/get_id works in isolation as expected."""

        f = Future(solver=None, id_=None)

        # f.id should be None
        self.assertIsNone(f.id)
        with self.assertRaises(TimeoutError):
            f.wait_id(timeout=1)

        # set it
        submission_id = 'test-id'
        f.id = submission_id

        # validate it's available
        self.assertEqual(f.wait_id(), submission_id)
        self.assertEqual(f.wait_id(timeout=1), submission_id)
        self.assertEqual(f.id, submission_id)

    def test_id_integration(self):
        """Problem ID getter blocks correctly when ID set by the client."""

        submission_id = 'test-id'
        solver_name = 'solver-id'
        release_reply = threading.Event()

        # each thread can have its instance of a session because
        # we use a global lock (event) in the mocked responses
        def create_mock_session(client):
            session = mock.Mock()

            # delayed submit; emulates waiting in queue
            def post(path, _):
                release_reply.wait()
                reply_body = self.sapi.complete_reply(id=submission_id, solver=solver_name)
                return choose_reply(path, {'problems/': [reply_body]})

            session.post = post

            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem

                future = solver.sample_ising(linear, quadratic)

                # initially, the id is not available
                with self.assertRaises(TimeoutError):
                    future.wait_id(timeout=1)

                # release the mocked sapi reply with the id
                release_reply.set()

                # verify the id is now available
                self.assertEqual(future.wait_id(), submission_id)

    @unittest.skipUnless(dimod, "dimod required for 'Solver.sample_bqm'")
    def test_sampleset_id(self):
        f = Future(solver=None, id_=None)

        # f.id should be None
        self.assertIsNone(f.id)
        with self.assertRaises(TimeoutError):
            f.sampleset.wait_id(timeout=1)

        # set it
        submission_id = 'test-id'
        f.id = submission_id

        # validate it's available
        self.assertEqual(f.sampleset.wait_id(), submission_id)
        self.assertEqual(f.sampleset.wait_id(timeout=1), submission_id)

@mock.patch('time.sleep', lambda *x: None)
class TestOffsetHandling(_QueryTest):

    @classmethod
    def setUpClass(cls):
        cls.sapi = StructuredSapiMockResponses()

    def test_submit_offset_answer_includes_it(self):
        """Handle a normal query with offset and response that includes it."""

        # ising problem energy offset
        offset = 3

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(
                    id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(
                    id='123', answer_patch=dict(offset=offset))})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)
                results = solver.sample_ising(linear, quadratic, offset, **params)

                self._check(results, linear, quadratic, offset=offset, **params)

    def test_submit_offset_answer_does_not_include_it(self):
        """Handle a normal query with offset and response that doesn't include it."""

        # ising problem energy offset
        offset = 3

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(id='123')})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)
                results = solver.sample_ising(linear, quadratic, offset, **params)

                # although SAPI response doesn't include offset, Future should patch it on-the-fly
                self._check(results, linear, quadratic, offset=offset, **params)

    def test_submit_offset_wrong_offset_in_answer(self):
        """Energy levels don't match because offset in answer is respected, even if wrong"""

        # ising problem energy offset
        offset = 3
        answer_offset = 2 * offset      # make it wrong

        # each thread can have its instance of a session because
        # the mocked responses are stateless
        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(
                    id='123', answer_patch=dict(offset=answer_offset))})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)
                results = solver.sample_ising(linear, quadratic, offset, **params)

                # since SAPI response includes offset, Future shouldn't patch it;
                # but because the offset in answer is wrong, energies are off
                with self.assertRaises(AssertionError):
                    self._check(results, linear, quadratic, offset=offset, **params)


@mock.patch('time.sleep', lambda *x: None)
class TestComputationDeprecations(_QueryTest):

    @classmethod
    def setUpClass(cls):
        cls.sapi = StructuredSapiMockResponses()

    def test_deprecations(self):
        """Proper deprecation warnings are raised."""

        def create_mock_session(client):
            session = mock.Mock()
            session.post = lambda a, _: choose_reply(a, {
                'problems/': [self.sapi.complete_no_answer_reply(id='123')]})
            session.get = lambda a: choose_reply(a, {
                'problems/123/': self.sapi.complete_reply(id='123')})
            return session

        with mock.patch.object(Client, 'create_session', create_mock_session):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)

                linear, quadratic = self.sapi.problem
                params = dict(num_reads=100)
                results = solver.sample_ising(linear, quadratic, **params)

                # aliased keys are deprecated in 0.8.0
                with self.assertWarns(DeprecationWarning):
                    results['samples']
                with self.assertWarns(DeprecationWarning):
                    results['occurrences']


class TestProblemLabel(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.sapi = StructuredSapiMockResponses()

    class PrimaryAssertionSatisfied(Exception):
        """Raised by `on_submit_label_verifier` to signal correct label."""

    def on_submit_label_verifier(self, expected_label):
        """Factory for mock Client._submit() that will verify existence, and
        optionally validate label value.
        """

        # replacement for Client._submit()
        def _submit(client, body_data, computation):
            body = json.loads(body_data.result())

            if 'label' not in body:
                if expected_label is None:
                    raise TestProblemLabel.PrimaryAssertionSatisfied
                else:
                    raise AssertionError("label field missing")

            label = body['label']
            if label != expected_label:
                raise AssertionError(
                    "unexpected label value: {!r} != {!r}".format(label, expected_label))

            raise TestProblemLabel.PrimaryAssertionSatisfied

        return _submit

    def generate_sample_problems(self, solver):
        linear, quadratic = self.sapi.problem

        # test sample_{ising,qubo,bqm}
        problems = [("sample_ising", (linear, quadratic)),
                    ("sample_qubo", (quadratic,))]
        if dimod:
            bqm = dimod.BQM.from_ising(linear, quadratic)
            problems.append(("sample_bqm", (bqm,)))

        return problems

    @parameterized.expand([
        ("undefined", None),
        ("empty", ""),
        ("string", "text label")
    ])
    @mock.patch.object(Client, 'create_session', lambda client: mock.Mock())
    def test_label_is_sent(self, name, label):
        """Problem label is set on problem submit."""

        with Client(endpoint='endpoint', token='token') as client:
            solver = Solver(client, self.sapi.solver.data)
            problems = self.generate_sample_problems(solver)

            for method_name, problem_args in problems:
                with self.subTest(method_name=method_name):
                    sample = getattr(solver, method_name)

                    with mock.patch.object(Client, '_submit', self.on_submit_label_verifier(label)):

                        with self.assertRaises(self.PrimaryAssertionSatisfied):
                            sample(*problem_args, label=label).result()

    @parameterized.expand([
        ("undefined", None),
        ("empty", ""),
        ("string", "text label")
    ])
    def test_label_is_received(self, name, label):
        """Problem label is set from response in result/sampleset."""

        def make_session_generator(label):
            def create_mock_session(client):
                session = mock.Mock()
                session.post = lambda a, _: choose_reply(a, {
                    'problems/': [self.sapi.complete_no_answer_reply(id='123', label=None)]})
                session.get = lambda a: choose_reply(a, {
                    'problems/123/': self.sapi.complete_reply(id='123', label=label)})
                return session
            return create_mock_session

        with mock.patch.object(Client, 'create_session', make_session_generator(label)):
            with Client(endpoint='endpoint', token='token') as client:
                solver = Solver(client, self.sapi.solver.data)
                problems = self.generate_sample_problems(solver)

                for method_name, problem_args in problems:
                    with self.subTest(method_name=method_name):
                        sample = getattr(solver, method_name)

                        future = sample(*problem_args, label=label)
                        future.result()     # ensure future is resolved

                        self.assertEqual(future.label, label)

                        # sampleset will only be available if dimod is installed
                        if dimod:
                            info = future.sampleset.info
                            self.assertEqual(info.get('problem_label'), label)


class TestSerialization(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.sapi = StructuredSapiMockResponses()

    class AssertionSatisfied(Exception):
        """Raised by `on_submit_data_verifier` to signal correct serialization."""

    def on_submit_data_verifier(self, expected_params):
        """Factory for mock Client._submit() that will validate parameter values."""

        # replacement for Client._submit(), called with exact network request data
        def _submit(client, body_data, computation):
            body = json.loads(body_data.result())

            params = body.get('params')
            if params != expected_params:
                raise AssertionError("params don't match")

            raise TestSerialization.AssertionSatisfied

        return _submit

    def generate_sample_problems(self, solver):
        linear, quadratic = self.sapi.problem

        # test sample_{ising,qubo,bqm}
        problems = [("sample_ising", (linear, quadratic)),
                    ("sample_qubo", (quadratic,))]
        if dimod:
            bqm = dimod.BQM.from_ising(linear, quadratic)
            problems.append(("sample_bqm", (bqm,)))

        return problems

    @parameterized.expand([
        (numpy.bool_(1), True), (numpy.bool8(1), True),
        (numpy.byte(1), 1), (numpy.int8(1), 1),
        (numpy.ubyte(1), 1), (numpy.uint8(1), 1),
        (numpy.short(1), 1), (numpy.int16(1), 1),
        (numpy.ushort(1), 1), (numpy.uint16(1), 1),
        (numpy.intc(1), 1), (numpy.int32(1), 1),
        (numpy.uintc(1), 1), (numpy.uint32(1), 1),
        (numpy.int_(1), 1), (numpy.int32(1), 1),
        (numpy.uint(1), 1), (numpy.uint32(1), 1),
        (numpy.longlong(1), 1), (numpy.int64(1), 1),
        (numpy.ulonglong(1), 1), (numpy.uint64(1), 1),
        (numpy.half(1.0), 1.0), (numpy.float16(1.0), 1.0),
        (numpy.single(1.0), 1.0), (numpy.float32(1.0), 1.0),
        (numpy.double(1.0), 1.0), (numpy.float64(1.0), 1.0),
        (numpy.longdouble(1.0), 1.0)
    ] + ([
        (numpy.float128(1.0), 1.0)      # unavailable on windows
    ] if hasattr(numpy, 'float128') else [
    ]))
    @mock.patch.object(Client, 'create_session', lambda client: mock.Mock())
    def test_params_are_serialized(self, np_val, py_val):
        """Parameters supplied as NumPy types are correctly serialized."""

        user_params = dict(num_reads=np_val)
        expected_params = dict(num_reads=py_val)

        with Client(endpoint='endpoint', token='token') as client:
            solver = Solver(client, self.sapi.solver.data)
            problems = self.generate_sample_problems(solver)

            for method_name, problem_args in problems:
                with self.subTest(method_name=method_name, np_val=np_val, py_val=py_val):
                    sample = getattr(solver, method_name)

                    with mock.patch.object(
                            Client, '_submit', self.on_submit_data_verifier(expected_params)):

                        with self.assertRaises(self.AssertionSatisfied):
                            sample(*problem_args, **user_params).result()
